#!/usr/bin/env python3
"""
https://www.ebi.ac.uk/gwas/rest/docs/api
"""
###
import sys,os,re,json,time,logging,tqdm
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
import urllib.request,urllib.parse
import pandas as pd
#
API_HOST='www.ebi.ac.uk'
API_BASE_PATH='/gwas/rest/api'
BASE_URL='https://'+API_HOST+API_BASE_PATH
#
NCHUNK=100;
#
##############################################################################
def ListStudies(base_url=BASE_URL, fout=None):
  """Only simple metadata."""
  tags=[]; n_study=0; rval=None; df=None; tq=None;

  retry_strategy = Retry(
	total=10,
	backoff_factor=2,
	status_forcelist=[413, 429, 500, 502, 503, 504],
	method_whitelist=["HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE"]
	)
  adapter = HTTPAdapter(max_retries=retry_strategy)
  session = requests.Session()
  session.mount("https://", adapter)
  session.mount("http://", adapter)

  url_this = base_url+f'/studies?size={NCHUNK}'
  while True:
    logging.debug(url_this)
    response = session.get(url_this)
    if (response.status_code!=200):
      logging.error(f"(status_code={response.status_code}): url_this: {url_this}")
      break
    try:
      rval = response.json()
    except Exception as e:
      logging.error(f"{str(e)}; response.text: {response.text}")
      break
    if '_embedded' not in rval or 'studies' not in rval['_embedded']: break
    studies = rval['_embedded']['studies']
    if studies is None: break
    if tq is None: tq = tqdm.tqdm(total=rval["page"]["totalElements"])
    for study in studies:
      tq.update(n=1)
      if not tags:
        tags = list(study.keys())
        for tag in tags[:]:
          if type(study[tag]) in (list, dict) and tag!="diseaseTrait":
            tags.remove(tag)
            logging.info(f"Ignoring tag: {tag}")
      df_this = pd.DataFrame({tag:[str(study[tag]) if tag in study else ''] for tag in tags})
      if fout is not None: df_this.to_csv(fout, "\t", index=False, header=(n_study==0), mode=('w' if n_study==0 else 'a'))
      else: df = pd.concat([df, df_this])
      n_study+=1
    if 'next' not in rval['_links']: break
    elif url_this == rval['_links']['last']['href']: break
    else: url_this = rval['_links']['next']['href']
  logging.info(f"n_study: {n_study}")
  return(df)

##############################################################################
def GetStudyAssociations(ids, skip=0, nmax=None, base_url=BASE_URL, fout=None):
  """
Mapped genes via SNP links.
arg = authorReportedGene
sra = strongestRiskAllele
https://www.ebi.ac.uk/gwas/rest/api/studies/GCST001430/associations?projection=associationByStudy
  """
  n_id=0; n_assn=0; n_loci=0; n_arg=0; n_sra=0; n_snp=0; df=None; tq=None;
  quiet = bool(logging.getLogger().getEffectiveLevel()>15)
  gcsts=set([]); tags_assn=[]; tags_study=[]; tags_locus=[]; tags_sra=[]; tags_arg=[];
  url = base_url+'/studies'
  if skip>0: logging.info(f"SKIP IDs skipped: {skip}")
  for id_this in ids[skip:]:
    if not quiet and tq is None: tq = tqdm.tqdm(total=len(ids)-skip)
    if tq is not None: tq.update()
    url_this = url+f'/{id_this}/associations?projection=associationByStudy'
    response = requests.get(url_this)
    if (response.status_code!=200):
      logging.error(f"(status_code={response.status_code}): url_this: {url_this}")
      continue
    rval = response.json()
    if not rval: continue
    if '_embedded' in rval and 'associations' in rval['_embedded']:
      assns = rval['_embedded']['associations']
    else:
      logging.error(f'No associations for study: {id_this}')
      continue
    df_this=None;
    for assn in assns:
      n_assn+=1
      if n_assn==1:
        for key,val in assn.items():
          if type(val) not in (list, dict):
            tags_assn.append(key)
        for key in assn['study'].keys():
          if type(assn['study'][key]) not in (list, dict):
            tags_study.append(key)
        for key in assn['loci'][0].keys():
          if key not in ('strongestRiskAlleles', 'authorReportedGenes'):
            tags_locus.append(key)
        for key in assn['loci'][0]['strongestRiskAlleles'][0].keys():
          if key != '_links':
            tags_sra.append(key)
      df_assn = pd.DataFrame({tag_assn:[assn[tag_assn]] for tag_assn in tags_assn})
      df_study = pd.DataFrame({tag_study:[assn['study'][tag_study]] for tag_study in tags_study})
      df_locus = pd.DataFrame({tag_locus:[assn['loci'][0][tag_locus]] for tag_locus in tags_locus})
      df_locus.columns = ['locus_'+s for s in df_locus.columns]
      df_sra = pd.DataFrame({tag_sra:[assn['loci'][0]['strongestRiskAlleles'][0][tag_sra]] for tag_sra in tags_sra})
      df_sra.columns = ['allele_'+s for s in df_sra.columns]
      df_assn = pd.concat([df_assn, df_study, df_locus, df_sra], axis=1)
      df_this = pd.concat([df_this, df_assn], axis=0)
      if 'accessionId' in assn['study']: gcsts.add(assn['study']['accessionId'])
      n_loci += len(assn['loci'])
      for locus in assn['loci']:
        n_sra += len(locus['strongestRiskAlleles'])
        for sra in locus['strongestRiskAlleles']:
          snp_href = sra['_links']['snp']['href'] if '_links' in sra and 'snp' in sra['_links'] and 'href' in sra['_links']['snp'] else ''
          if snp_href: n_snp+=1
    if fout: df_this.to_csv(fout, "\t", index=False, header=(n_id==0), mode=('w' if n_id==0 else 'a'))
    if fout is None: df = pd.concat([df, df_this], axis=0)
    n_id+=1
    if n_id==nmax:
      logging.info(f"NMAX IDs reached: {nmax}")
      break
  if tq is not None: tq.close()
  n_gcst = len(gcsts)
  logging.info(f"INPUT RCSTs: {n_id}; OUTPUT RCSTs: {n_gcst} ; assns: {n_assn} ; loci: {n_loci} ; alleles: {n_sra} ; snps: {n_snp}")
  if fout is None: return(df)

##############################################################################
def GetSnps(ids, skip=0, nmax=None, base_url=BASE_URL, fout=None):
  """
Input: rs_id, e.g. rs7329174
loc = location
gc = genomicContext
  """
  n_snp=0; n_gene=0; n_loc=0; df=None; tq=None;
  tags_snp=[]; tags_loc=[]; tags_gc=[]; tags_gcloc=[];  tags_gene=[]; 
  quiet = bool(logging.getLogger().getEffectiveLevel()>15)

  retry_strategy = Retry(
	total=10,
	backoff_factor=2,
	status_forcelist=[413, 429, 502, 503, 504], #Not 500
	method_whitelist=["HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE"]
	)
  adapter = HTTPAdapter(max_retries=retry_strategy)
  session = requests.Session()
  session.mount("https://", adapter)
  session.mount("http://", adapter)

  url = base_url+'/singleNucleotidePolymorphisms'
  if skip>0: logging.info(f"SKIP IDs skipped: {skip}")
  for id_this in ids[skip:]:
    if not quiet and tq is None: tq = tqdm.tqdm(total=len(ids)-skip)
    if tq is not None: tq.update()
    url_this = url+'/'+id_this
    #response = requests.get(url_this)
    try:
      response = session.get(url_this)
    except Exception as e:
      logging.error(f"Failed for SNP:{id_this}; URL:{url_this}; {str(e)}")
      continue
    if response.status_code!=200:
      logging.error(f"(status_code={response.status_code}): url_this: {url_this}")
      continue
    snp = response.json()
    if 'genomicContexts' not in snp: continue
    if len(snp['genomicContexts'])==0: continue
    df_this=None;
    for gc in snp['genomicContexts']:
      if not tags_snp:
        for key,val in snp.items():
          if type(val) not in (list, dict): tags_snp.append(key)
        for key in gc.keys():
          if key not in ('gene', 'location', '_links'): tags_gc.append(key)
        for key in gc['location'].keys():
          if key != '_links': tags_gcloc.append(key)
        for key in gc['gene'].keys():
          if key != '_links': tags_gene.append(key)
      df_snp = pd.DataFrame({tag_snp:[snp[tag_snp]] for tag_snp in tags_snp})
      df_gc = pd.DataFrame({tag_gc:[gc[tag_gc]] for tag_gc in tags_gc})
      gcloc = gc['location']
      df_gcloc = pd.DataFrame({tag_gcloc:[gcloc[tag_gcloc]] for tag_gcloc in tags_gcloc})
      gene = gc['gene']
      try: gene["ensemblGeneIds"] = (",".join([gid["ensemblGeneId"] for gid in gene["ensemblGeneIds"]]))
      except: pass
      try: gene["entrezGeneIds"] = (",".join([gid["entrezGeneId"] for gid in gene["entrezGeneIds"]]))
      except: pass
      df_gene = pd.DataFrame({tag_gene:[gene[tag_gene]] for tag_gene in tags_gene})
      df_snp = pd.concat([df_snp, df_gc, df_gcloc, df_gene], axis=1)
      df_this = pd.concat([df_this, df_snp], axis=0)
      n_gene+=1
    if tq is not None: tq.close()
    if fout: df_this.to_csv(fout, "\t", index=False, header=(n_snp==0), mode=('w' if n_snp==0 else 'a'))
    if fout is None: df = pd.concat([df, df_this], axis=0)
    n_snp+=1
    if n_snp==nmax:
      logging.info(f"NMAX IDs reached: {nmax}")
      break
  logging.info(f"SNPs: {n_snp}; genes: {n_gene}")
  if fout is None: return(df)

##############################################################################
def SearchStudies(ids, searchtype, base_url=BASE_URL, fout=None):
  tags=[]; n_study=0; rval=None; df=None; tq=None;
  url = base_url+'/studies/search'
  if searchtype=='gcst':
    url+='/'
  elif searchtype.lower()=='pubmedid':
    url+='/findByPublicationIdPubmedId?pubmedId={}'
  elif searchtype.lower()=='efotrait':
    url+='/findByEfoTrait?efoTrait={}'
  elif searchtype.lower()=='efouri':
    url+='/findByEfoUri?efoUri={}'
  elif searchtype.lower()=='accessionid':
    url+='/findByAccessionId?accessionId={}'
  else:
    logging.error(f'Searchtype not supported: {searchtype}')
    return
  for id_this in ids:
    url_this = url.format(urllib.parse.quote(id_this))
    response = requests.get(url_this)
    if (response.status_code!=200):
      logging.error(f"(status_code={response.status_code}): url_this: {url_this}")
      break
    rval = response.json()
    if not rval or '_embedded' not in rval or 'studies' not in rval['_embedded']: continue
    studies = rval['_embedded']['studies']
    if not studies: continue
    for study in studies:
      if not tags:
        for tag in study.keys():
          if type(study[tag]) not in (list, dict) or tag=="diseaseTrait":
            tags.append(tag) #Only simple metadata.
      n_study+=1
      df_this = pd.DataFrame({tag:[str(study[tag]) if tag in study else ''] for tag in tags})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, "\t", index=False)
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
  logging.info(f"n_study: {n_study}")
  if fout is None: return(df)

##############################################################################



#!/usr/bin/env python3
"""
Pharos  REST API client
https://pharos.nih.gov/idg/api/v1/targets(589)

THIS PROBABLY OBSOLETED BY GRAPHQL API/CLIENT; SEE pharos/Utils.py.
"""
###
import sys,os,json,re,time,logging
#
from ..util import rest
#
#############################################################################
def GetTargets(base_url, ids, idtype, fout):
  tags=[]; n_out=0;
  for id_this in ids:
    url = base_url+'/targets(%s)'%id_this
    rval = rest.GetURL(url, parse_json=True)
    if not rval:
      logging.debug('Not found: %s'%(id_this))
      continue
    tgt = rval
    if not tags:
      for tag in tgt.keys():
        if not tag.startswith('_') and (type(tgt[tag]) not in (list, dict)):
          tags.append(tag)
      fout.write('\t'.join(tags)+'\n')
    vals = [(str(tgt[tag]) if tag in tgt else '') for tag in tags]
    fout.write('\t'.join(vals)+'\n')
    n_out+=1
  logging.info('n_in: %d; n_out: %d'%(len(ids), n_out))

#############################################################################
def GetTargetProperties(base_url, ids, idtype, fout):
  tags=[]; n_out=0;
  for id_this in ids:
    url = (base_url+'/targets(%s)/properties'%id_this)
    rval = rest.GetURL(url, parse_json=True)
    if not rval:
      logging.debug('Not found: %s'%(id_this))
      continue
    props = rval
    for prop in props:
      if not tags:
        for tag in prop.keys():
          if not tag.startswith('_') and (type(prop[tag]) not in (list, dict)):
            tags.append(tag)
        fout.write('\t'.join([idtype]+tags)+'\n')
      vals = [id_this]+[(str(prop[tag]) if tag in prop else '') for tag in tags]
      fout.write('\t'.join(vals)+'\n')
      n_out+=1
  logging.info('n_in: %d; n_out: %d'%(len(ids), n_out))

#############################################################################
def ListItems(mode, base_url, fout):
  n_out=0; tags=[]; top=100; skip=0;
  while True:
    url = base_url+'/%s?top=%d&skip=%d'%(mode, top, skip)
    rval = rest.GetURL(url, parse_json=True)
    if not rval:
      break
    elif type(rval) is not dict:
      logging.error('rval="%s"'%(str(rval)))
      break
    logging.debug(json.dumps(rval, indent=2))
    count = rval['count'] if 'count' in rval else None
    total = rval['total'] if 'total' in rval else None
    uri = rval['uri'] if 'uri' in rval else None
    logging.debug('uri="%s"'%(uri))
    items = rval['content']
    if not items: break
    for item in items:
      if not tags:
        for tag in item.keys():
          if not tag.startswith('_') and (type(item[tag]) not in (list, dict)):
            tags.append(tag)
        fout.write('\t'.join(tags)+'\n')
      vals = [(str(item[tag]) if tag in item else '') for tag in tags]
      fout.write('\t'.join(vals)+'\n')
      n_out+=1
    skip+=top
  logging.info('n_out = %d'%(n_out))

#############################################################################



#!/usr/bin/env python3
"""
Utility functions for BioGRID REST API.
See: http://wiki.thebiogrid.org/doku.php/biogridrest
"""
###
import sys,os,re,json,time,logging,yaml
#
from ..util import rest
#
##############################################################################
def ReadParamFile(fparam):
  params={};
  with open(fparam, 'r') as fh:
    for param in yaml.load_all(fh, Loader=yaml.BaseLoader):
      for k,v in param.items():
        params[k] = v
  return params

##############################################################################
def ListOrganisms(base_url, params, fout):
  n_all=0; n_out=0; n_err=0;
  url=base_url+'/organisms/?accesskey=%s'%params['API_KEY']
  url+=('&format=tab2')
  rval=rest.GetURL(url, parse_json=False)
  txt = rval
  lines = txt.splitlines()
  fout.write('organism_id, organism"\n')
  for line in lines:
    oid,org = re.split('\t', line)
    fout.write('%s\t"%s"\n'%(oid, org))
    n_out+=1
  logging.info('n_out: %d'%(n_out))

##############################################################################
def ListIdTypes(base_url, params, fout):
  n_all=0; n_out=0; n_err=0;
  url=base_url+'/identifiers/?accesskey=%s'%params['API_KEY']
  url+=('&format=tab2')
  rval=rest.GetURL(url, parse_json=False)
  txt = rval
  lines = txt.splitlines()
  for line in lines:
    fout.write('%s\n'%(line))
    n_all+=1
    n_out+=1
  logging.info('n_out: %d'%(n_out))

##############################################################################
def GetInteractions(base_url, params, ids, fout):
  n_all=0; n_out=0; n_err=0;
  t0=time.time()
  tags=[];

  for iid in ids:
    url=base_url+'/interactions/%s?'%iid
    url+=('&accesskey=%s&format=json'%params['API_KEY'])
    rval=rest.GetURL(url, parse_json=True)
    logging.debug(json.dumps(rval, indent=2, sort_keys=False)+'\n')

    if type(rval) is not dict:
      n_err+=1
      continue
    if not iid in rval :
      n_err+=1
      continue
    intr = rval[iid]
    n_all+=1

    if n_all==1 or not tags:
      tags=intr.keys()
      fout.write('\t'.join(tags)+'\n')

    vals=[intr[tag] if tag in intr else '' for tag in tags]
    fout.write(('\t'.join(vals))+'\n')
    n_out+=1

  logging.info('n_all: %d; n_out: %d; n_err: %d'%(n_all, n_out, n_err))

##############################################################################
def SearchInteractions(base_url, params, ids, search_params, fout):
  n_all=0; n_out=0; n_err=0;
  t0=time.time()
  tags=[];

  url=base_url+('/interactions/?accesskey=%s&format=json'%params['API_KEY'])
  if ids:
    url+=('&geneList=%s'%('|'.join(ids)))
  url+=('&interSpeciesExcluded=%s'%str(not search_params['inc_interspecies']).lower())
  url+=('&selfInteractionsExcluded=%s'%str(not search_params['inc_self']).lower())
  if search_params['elist']:
    url+=('&includeEvidence=%s'%str(search_params['inc_evidence']).lower())
    url+=('&evidenceList=%s'%('|'.join(search_params['elist'])))
  if search_params['addl_idtypes']:
    url+=('&additionalIdentifierTypes=%s'%('|'.join(search_params['addl_idtypes'])))
  if search_params['human']:
    url+=('&taxId=9606')

  skip=0; chunk=1000;

  while True:
    url_this=url+('&start=%d&max=%d'%(skip, chunk))
    rval=rest.GetURL(url_this, parse_json=True)
    logging.debug(json.dumps(rval, indent=2, sort_keys=False)+'\n')
    if not rval: break
    if type(rval) is not dict:
      n_err+=1
      continue
    intrs = rval
    for iid, intr in intrs.items():
      n_all+=1
      if n_all==1 or not tags:
        tags=intr.keys()
        fout.write('\t'.join(tags)+'\n')
      vals=[];
      for tag in tags:
        vals.append(intr[tag] if tag in intr else '')
      fout.write(('\t'.join(vals))+'\n')
      n_out+=1
    skip+=chunk

  logging.info('n_all: %d; n_out: %d; n_err: %d'%(n_all, n_out, n_err))
  logging.info(('Elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0)))))

##############################################################################



#!/usr/bin/env python
#############################################################################
### pubchem_compound_assaystats.py - From PubChem CSV assay files, determine
### compound activity stats.
### activity outcomes: 
###   1 = inactive
###   2 = active
###   3 = inconclusive
###   4 = unspecified
###   multiple, differing 1, 2 or 3 = discrepant
###   not 4 = tested
###
#############################################################################
### This program is a memory hog (700MB for 380k mols).  Use with care.
#############################################################################
###     aTested - how many assays where that cpd has been tested
###     aActive - how many assays where that cpd has been tested active
###     sTested - how many samples with that cpd have been tested
###     sActive - how many samples with that cpd have been tested active
#############################################################################
### Jeremy Yang
###  5 Jul 2012
#############################################################################
import os,sys,re,getopt,gzip,zipfile

import hier_scaffolds_assaystats  ## for ExtractOutcomes()

#############################################################################
def main():
  global PROG,SCRATCHDIR
  PROG=os.path.basename(sys.argv[0])
  SCRATCHDIR=os.getcwd()+'/data/scratch'

  usage='''
  %(PROG)s - 

  required:
  --inmols=<INMOLS> ... mols w/ cids (smiles)
  --csvdir=<DIR>   ... dir containing input csv.gz assay data
  --zipdir=<DIR>   ... dir containing zipfiles containing input csv.gz assay data
  --csvfile=<FILE>   ... input csv.gz assay data
  --o=<OUTSCAFS> ... output mols with data (smiles)

  options:
  --aids=<AIDS> ... list of AIDs to select within --csvdir dir
  --aidfile=<AIDFILE> ... file of AIDs to select within --csvdir dir
  --cidfile=<CIDFILE> ... file of CIDs to select (ignore others)

  --use_sids     ... use SIDs instead of CIDs (all i/o)
  --n_max_aids=<N> ... mostly for debugging
  --v            ... verbose
  --vv           ... very verbose
  --h            ... this help
'''%{'PROG':PROG}

  def ErrorExit(msg):
    print >>sys.stderr,msg
    sys.exit(1)

  #csvdir="/home/data/pubchem/bioassay/csv/data";
  csvdir=None
  zipdir="/home/data/pubchem/bioassay/csv/data";
  csvfile=None; verbose=0;
  ofile=None; inmolsfile=None;
  aids=None; aidfile=None; cidfile=None;
  use_sids=False;
  n_max_aids=None;
  opts,pargs = getopt.getopt(sys.argv[1:],'',['h','v','vv','csvdir=','zipdir=',
  'aids=','aidfile=','csvfile=','o=','inmols=','n_max_aids=',
  'cidfile=','use_sids'])
  if not opts: ErrorExit(usage)
  for (opt,val) in opts:
    if opt=='--h': ErrorExit(usage)
    elif opt=='--inmols': inmolsfile=val
    elif opt=='--o': ofile=val
    elif opt=='--csvdir': csvdir=val
    elif opt=='--zipdir': zipdir=val
    elif opt=='--csvfile': csvfile=val
    elif opt=='--aids': aids=val
    elif opt=='--aidfile': aidfile=val
    elif opt=='--cidfile': cidfile=val
    elif opt=='--use_sids': use_sids=True
    elif opt=='--n_max_aids': n_max_aids=int(val)
    elif opt=='--vv': verbose=2
    elif opt=='--v': verbose=1
    else: ErrorExit('Illegal option: %s'%val)

  if not (csvdir or zipdir or csvfile):
    ErrorExit('csvdir or zipdir or file required\n'+usage)
  if not ofile:
    ErrorExit('output file required\n'+usage)
  if not (inmolsfile):
    ErrorExit('--inmols file required\n'+usage)

  finmols=file(inmolsfile)
  if not finmols: ErrorExit('cannot open: %s'%inmolsfile)
  fout=file(ofile,'w')
  if not fout: ErrorExit('cannot open: %s'%ofile)

  aidset=[]
  if aids:
    for val in re.split('[\s,]+',aids):
      if not val.strip(): continue
      aidset.append(int(val.strip()))
  if aidfile:
    f=file(aidfile)
    if not f:
      ErrorExit('cannot open: %s'%aidfile)
    lines=f.readlines()
    for line in lines:
      if not line.strip(): continue
      aidset.append(int(line.strip()))
  cidset={}	## use a hash for speed (could be many cids)
  if cidfile:
    f=file(cidfile)
    if not f:
      ErrorExit('cannot open: %s'%cidfile)
    lines=f.readlines()
    for line in lines:
      if not line.strip(): continue
      line=re.sub('\s.*$','',line)
      cidset[int(line.strip())]=True
  if len(cidset.keys())==0: cidset=None

  files=[]
  if csvdir:
    for fname_csv_gz in os.listdir(csvdir):
      if not re.search('\.csv\.gz',fname_csv_gz): continue
      try:
        aid=int(re.sub('\.csv\.gz','',fname_csv_gz))
      except:
        print >>sys.stderr, 'cannot parse AID: "%s"'%fname_csv_gz
        continue
      if aidset and (aid not in aidset): continue
      fpath_csv_gz=csvdir+'/'+fname_csv_gz
      files.append((fpath_csv_gz,aid))
  elif zipdir:
    for fname_zip in os.listdir(zipdir):
      if not re.search('\.zip',fname_zip): continue
      fpath_zip=zipdir+'/'+fname_zip
      try:
        zf=zipfile.ZipFile(fpath_zip,'r')
      except:
        print >>sys.stderr, 'ERROR: cannot read fpath_zip: "%s"'%fpath_zip
        continue
      flist_csv_gz=zf.namelist()
      zf.close()
      for fpath_csv_gz in flist_csv_gz:
        if not re.search('\.csv\.gz',fpath_csv_gz): continue
        try:
          if re.search(r'/',fpath_csv_gz):
            txt=re.sub(r'^.*/(\d*)\.csv\.gz',r'\1',fpath_csv_gz)
          else:
            txt=re.sub(r'\.csv\.gz','',fpath_csv_gz)
          aid=int(txt)
        except:
          print >>sys.stderr, 'cannot parse AID: "%s"'%fpath_csv_gz
          print >>sys.stderr, 'DEBUG txt: "%s"'%txt
          continue
        if aidset and (aid not in aidset): continue
        files.append((fpath_zip,fpath_csv_gz,aid))
  else:
    try:
      aid=int(re.sub('\.csv\.gz','',os.path.basename(csvfile)))
    except:
      ErrorExit('cannot parse AID: "%s"'%csvfile)
    files.append((csvfile,aid))

  n_mols=0; cidlist=[]; cid2smi={}; 
  while True:
    line=finmols.readline()
    if not line: break
    n_mols+=1
    line=line.rstrip()
    fields=line.split()
    if len(fields)<3:
      print >>sys.stderr, 'Aaak! bad line:', line
      continue
    smiles=fields[0]
    cid=int(fields[1])
    cidlist.append(cid)
    cid2smi[cid]=smiles

  print >>sys.stderr, 'mols read: %d'%(n_mols,)

  if not (os.path.isdir(SCRATCHDIR) and os.access(SCRATCHDIR,os.W_OK)):
    try:
      os.mkdir(SCRATCHDIR)
    except:
      ErrorExit('SCRATCHDIR does not exist or is not writeable, and cannot be created: "%s"'%SCRATCHDIR)

  ### Read all [selected] assay csv files.  For each cid, create map of 
  ### aids and outcomes.
  ### (Non-active may be inactive, discrepant, etc.).
  ### Problem: if using sid, may be duplicates in a single assay, that is 
  ### multiple samples for one sid.
  cids={}; cids_active=[]; 
  for i_file,f in enumerate(files):
    if csvdir:
      fpath_csv,aid = f
      f_csv=gzip.open(fpath_csv)
    elif zipdir:
      fpath_zip,fpath_csv_gz,aid = f
      zf=zipfile.ZipFile(fpath_zip,'r')
      cwd=os.getcwd()
      os.chdir(SCRATCHDIR)
      zf.extract(fpath_csv_gz)
      os.chdir(cwd)
      zf.close()
      del zf #maybe free memory?
      f_csv=gzip.open(SCRATCHDIR+'/'+fpath_csv_gz)

    ftxt=f_csv.read()
    f_csv.close()
    del f_csv #maybe free memory?
    cids_this=hier_scaffolds_assaystats.ExtractOutcomes(ftxt,cidset,use_sids)
    del ftxt #maybe free memory?
    n_active=0;
    for cid in cids_this.keys():
      if not cid2smi.has_key(cid): continue
      if not cids.has_key(cid): cids[cid]={}
      cids[cid][aid]=cids_this[cid]['outcome']
      if cids[cid][aid]==2: n_active+=1
    if verbose:
      sys.stderr.write('%3d. AID %4d: '%(i_file,aid))
      sys.stderr.write('active/total: %6d /%6d\n'%(n_active,len(cids_this.keys())))
      if not i_file%10:
        sys.stderr.write('done %3d / %3d (%.1f%%)'%(i_file,len(files),100.0*i_file/len(files)))
    del cids_this #maybe free memory?

    os.unlink(SCRATCHDIR+'/'+fpath_csv_gz)


    if n_max_aids and i_file==n_max_aids:
      print >>sys.stderr, 'n_max_aids limit reached: %d'%n_max_aids
      break

  print >>sys.stderr, 'assay files read: %d'%i_file

  n_cid_notfound=0
  fout.write("#smiles cid aTested aActive sTested sActive\n")
  ### For each cid, generate stats.
  for cid in cidlist:
    aids_tested=[]; aids_active=[];
    n_samples=0; n_samples_active=0;
    smiles=cid2smi[cid]

    if not cids.has_key(cid):
      print >>sys.stderr, 'NOTE: cannot find cid %d in any assay.'%(cid)
      n_cid_notfound+=1
      continue

    for aid in cids[cid].keys():
      if aid not in aids_tested: aids_tested.append(aid)
      n_samples+=1
      if cids[cid][aid]==2:	#outcome
        n_samples_active+=1
        if aid not in aids_active: aids_active.append(aid)

    fout.write("%s %d %d %d %d %d\n"%(
	smiles,
	cid,
	len(aids_tested),
	len(aids_active),
	n_samples,
	n_samples_active))
    if verbose>1:
      sys.stderr.write(
	"cid=%d,aTested=%d,aActive=%d,sTested=%d,sActive%d\n"%(
	cid,
	len(aids_tested),
	len(aids_active),
	n_samples,
	n_samples_active))

  fout.close()

  id='CID'
  if use_sids: id='SID'
  print >>sys.stderr, '%s: number of assay files: %d'%(PROG,len(files))
  print >>sys.stderr, '%s: total %ss: %d'%(PROG,id,len(cids.keys()))
  print >>sys.stderr, '%s: number of %ss: %d'%(PROG,id,len(cidlist))
  print >>sys.stderr, '%s: number of %ss not found in any assay: %d'%(PROG,id,n_cid_notfound)

#############################################################################
if __name__=='__main__':
  #import cProfile
  #cProfile.run('main()')
  main()




#!/usr/bin/env python3
"""
OpenFDA Adverse Event Reports REST API client.
https://open.fda.gov/apis/
"""
### 
import sys,os,re,json,argparse,time,yaml,logging

from ... import fda
#
#############################################################################
def ReadParamFile(fparam):
  params={};
  with open(fparam, 'r') as fh:
    for param in yaml.load_all(fh, Loader=yaml.BaseLoader):
      for k,v in param.items():
        params[k] = v
  return params

##############################################################################
if __name__=='__main__':
  NMAX=100;
  epilog='''\
Example UNII: 786Z46389E
'''
  parser = argparse.ArgumentParser(description='OpenFDA Adverse Event Reports client', epilog=epilog)
  ops = ['search', 'counts', 'info', 'showfields']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--drug_class", help="EPC pharmacologic class")
  parser.add_argument("--drug_ind", help="drug indication")
  parser.add_argument("--drug_unii", help="drug ID UNII")
  parser.add_argument("--drug_ndc", help="drug ID NDC")
  parser.add_argument("--drug_spl", help="drug ID SPL")
  parser.add_argument("--serious", type=bool, help="serious adverse events")
  parser.add_argument("--fatal", type=bool, help="fatal adverse events (seriousnessdeath)")
  parser.add_argument("--tfrom", help="time-from (received by FDA) (YYYYMMDD)")
  parser.add_argument("--tto", default=time.strftime('%Y%m%d',time.localtime()), help="time-to (received by FDA) (YYYYMMDD)")
  parser.add_argument("--rawquery")
  parser.add_argument("--nmax", type=int, default=NMAX, help="max returned records")
  parser.add_argument("--api_host", default=fda.aer.API_HOST)
  parser.add_argument("--api_base_path", default=fda.aer.API_BASE_PATH)
  parser.add_argument("--param_file", default=os.environ['HOME']+"/.fda.yaml")
  parser.add_argument("--api_key")
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  api_base_url = 'https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  params = ReadParamFile(args.param_file)
  if args.api_key: params['API_KEY'] = args.api_key
  if not params['API_KEY']:
    parser.error('Please specify valid API_KEY via --api_key or --param_file')

  t0=time.time()

  if args.op == "search":
    fda.aer.Utils.Search(args.drug_class, args.drug_ind, args.drug_unii, args.drug_ndc, args.drug_spl, args.tfrom, args.tto, args.serious, args.fatal, args.rawquery, args.nmax, params['API_KEY'], api_base_url, fout)

  elif args.op == "info":
    rval = fda.aer.Utils.Info(api_base_url)
    for field in rval.keys():
      if field=='results': continue
      print(f"{field:16s}: {rval[field]}")

  elif args.op == "counts":
    print(fda.aer.Utils.GetCounts(args.tfrom, args.tto, api_base_url))

  elif args.op == "showfields":
    fields = fda.aer.Utils.GetFields(api_base_url)
    for field in fields:
      print(f"\t{field}")

  else:
    parser.error(f"Invalid operation: {args.op}")

  logging.info(f"Elapsed time: {time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0))}")



#!/usr/bin/env python3
"""
IDG Resource Submission System
https://rss.ccs.miami.edu/
"""
###
import sys,os,re,json,time,logging,tqdm
import pandas as pd
import requests
import urllib,urllib.parse
#
#
API_HOST="rss.ccs.miami.edu"
API_BASE_PATH="/rss-api"
BASE_URL = 'https://'+API_HOST+API_BASE_PATH
#
##############################################################################
def ListTargets(base_url=BASE_URL, fout=None):
  tags=[]; tq=None; df=pd.DataFrame(); 
  url = (base_url+f'/target')
  resp = requests.get(url, verify=False)
  targets = resp.json() if resp.status_code==200 else []
  for target in targets:
    logging.debug(json.dumps(target, sort_keys=True, indent=2))
    if not tags: tags = list(target.keys())
    df = pd.concat([df, pd.DataFrame({tags[j]:[target[tags[j]]] for j in range(len(tags))})])
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info("n_out: {}".format(df.shape[0]))
  return(df)

##############################################################################
def GetTargetResources(ids, base_url=BASE_URL, fout=None):
  tags=[]; tq=None; df=pd.DataFrame();
  for id_this in ids:
    url_this = (base_url+f'/target/id?id={id_this}')
    resp = requests.get(url_this, verify=False)
    resources = resp.json() if resp.status_code==200 else []
    for resource in resources:
      logging.debug(json.dumps(resource, sort_keys=True, indent=2))
      if not tags: tags = list(resource.keys())
      df = pd.concat([df, pd.DataFrame({tags[j]:[resource[tags[j]]] for j in range(len(tags))})])
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info("n_out: {}".format(df.shape[0]))
  return(df)

##############################################################################



#!/usr/bin/env python3
"""
PubChem PUG SOAP services, including:
  - PubChem Identifier Exchange Service (PIES), https://pubchem.ncbi.nlm.nih.gov/idexchange/idexchange.cgi
  - PubChem Standardization Service, https://pubchem.ncbi.nlm.nih.gov/standardize/standardize.cgi
  - PubChem Structure Search, https://pubchem.ncbi.nlm.nih.gov/search/search.cgi
See Save Query button for template XML.
https://pubchemdocs.ncbi.nlm.nih.gov/pug-soap
https://pubchemdocs.ncbi.nlm.nih.gov/pug-soap-reference
https://pubchem.ncbi.nlm.nih.gov/pug_soap/pug_soap.cgi?wsdl
"""
import sys,os,re,time,argparse,logging

from ... import pubchem

#############################################################################
if __name__=='__main__':
  SEARCH_NMAX=100;
  OPS = ["search_exact", "search_substructure", "search_similarity", "standardize", "idexchange"]
  OPERATORS = ["same", "parent", "samepar"]
  FORMATS = ["smiles", "smarts", "inchi", "inchikey", "cid", "sid", "sdf"]
  CHEMFORMATS = ["smiles", "inchi", "sdf"]
  SEARCHFORMATS = ["smiles", "smarts", "inchi", "cid"]
  epilog = """Search operations accept one query CID, SMILES, SMARTS, or InChI;
standardize and idexchange accept batch input files. Standardize accepts and returns
SMILES, SDF, or InChI. IDExchange accepts CID, SID, SMILES, InChI, InChIKey,
Synonym, and Registry ID (for specified Source)"""
  parser = argparse.ArgumentParser(description="PubChem PUG SOAP client: idexchange, standardize, sub|sim|exact search", epilog=epilog)
  parser.add_argument("op", choices=OPS, default="exact", help="OPERATION")
  parser.add_argument("--i", dest="ifile", help="Input file for batch IDs")
  parser.add_argument("--o", dest="ofile", help="Output SMILES|SDF|TSV file")
  parser.add_argument("--operator", choices=OPERATORS, default="same", help="Request logic operator")
  parser.add_argument("--ids", help="Input ID[s] (comma-separated)")
  parser.add_argument("--query_id", help="Input ID (use for search operations)")
  parser.add_argument("--ifmt", choices=FORMATS, default="smiles", help="Input format")
  parser.add_argument("--ofmt", choices=FORMATS, default="cid", help="Output format")
  parser.add_argument("--gz", action="store_true", help="Output gzipped [default via filename]")
  parser.add_argument("--sim_cutoff", type=float, default=pubchem.soap.SIM_CUTOFF, help="Similarity cutoff, Tanimoto")
  parser.add_argument("--active", help="Active mols only (in any assay)")
  parser.add_argument("--search_nmax", type=int, default=SEARCH_NMAX, help="Max search hits returned")
  parser.add_argument("--max_wait", type=int, default=pubchem.soap.MAX_WAIT, help="Max wait for query")
  parser.add_argument("--poll_wait", type=int, default=pubchem.soap.POLL_WAIT, help="Polling wait interval")
  parser.add_argument("--api_host", default=pubchem.soap.API_HOST)
  parser.add_argument("--api_base_path", default=pubchem.soap.API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  api_base_url = f"https://{args.api_host}{args.api_base_path}"

  do_gz = args.gz if args.gz else bool(args.ofile and args.ofile[-3:]=='.gz')
  fout = open(args.ofile, 'wb') if args.ofile and do_gz else open(args.ofile, 'w+') if args.ofile else sys.stdout
  ofmt = args.ofmt if args.ofmt else None

  if args.ofile and not args.ofmt:
    ofmt = re.sub('^.*\.', '', args.ofile[:-3]) if (args.ofile[-3:]=='.gz') else re.sub('^.*\.', '', args.ofile)

  ids=[];
  if args.query_id:
    ids = [args.query_id]
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids)
  elif args.ifile:
    with open(args.ifile) as fin:
      while True:
        line = fin.readline()
        if not line: break
        ids.append(line.rstrip())
  if len(ids)>0: logging.info(f"Input IDs: {len(ids)}")

  t0=time.time()	## start timer

  # Submit request (pubchem.soap.PugSoapRequest class):
  if args.op == "standardize":
    if args.ofmt not in CHEMFORMATS:
      parser.error(f"Output format must be in {str(CHEMFORMATS)} for operation {args.op}.")
    # TO DO: Loop for batch...
    pugreq = pubchem.soap.SubmitRequest_Standardize(ids[0], args.ifmt, args.ofmt, do_gz, api_base_url)
  elif args.op == "idexchange":
    pugreq = pubchem.soap.SubmitRequest_IDExchange(ids, args.ifmt, args.ofmt, args.operator, do_gz, api_base_url)
  elif args.op in ["search_exact", "search_substructure", "search_similarity"]:
    if args.ifmt not in SEARCHFORMATS:
      parser.error(f"Input format must be in {str(SEARCHFORMATS)} for operation {args.op}.")
    pugreq = pubchem.soap.SubmitRequest_StructuralSearch(ids[0], args.op, args.sim_cutoff, args.active, args.search_nmax, api_base_url)
  else:
    parser.error(f"Operation not suported: {args.op}")
    sys.exit(1)

  # Check request status.
  # Error status values include: "input-error", "server-error"
  if pugreq.status not in ("success", "queued", "running"):
    logging.error(f"Query failed; quitting (status={pugreq.status}, error={pugreq.error})")
    sys.exit()
  logging.info(f"Query [{pugreq.reqid}] status: {pugreq.status}")

  # Wait for result, success or failure.
  pugreq.waitForResult(args.poll_wait, args.max_wait)

  if pugreq.url is not None: # May have url for download.
    logging.info(f"url: {pugreq.url}")
    pubchem.soap.DownloadResults(pugreq.url, ofmt, do_gz, fout) # Output: download via url.
  elif pugreq.qkey is not None: # May have qkey for 2nd request.
    logging.info(f"qkey: {pugreq.qkey}")
    pugreq2 = pubchem.soap.MonitorQuery(pugreq.qkey, pugreq.wenv, ofmt, do_gz, api_base_url)
    pugreq2.waitForResult(args.poll_wait, args.max_wait)
    pubchem.soap.DownloadResults(pugreq2.url, ofmt, do_gz, fout) # Output: download via url.
  elif pugreq.out_struct is not None: # May have standardized structure.
    logging.info(f"out_struct: {pugreq.out_struct}")
    fout.write(f"{pugreq.out_struct}\n")
  else:
    logging.error(f"No url, qkey nor out_struct.")
  fout.close()

  logging.info(f"Output format: {ofmt}{(' (gzipped)' if do_gz else '')}")
  if args.ofile is not None: logging.info(f"Output file {args.ofile} ({os.stat(args.ofile).st_size/1e6:.2f}MB)")
  logging.info(f"Elapsed time: {time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0))}")




#!/usr/bin/env python3
"""
unichem_query.py - command line utility for UniChem REST API.
See: https://www.ebi.ac.uk/unichem/info/webservices

Source IDs:
1: ChEMBL
7: ChEBI
15: SureChEMBL
22: PubChem
"""
import sys,os,re,argparse,time,logging,json

from ..util import rest
#
PROG=os.path.basename(sys.argv[0])
#
API_HOST='www.ebi.ac.uk'
API_BASE_PATH='/unichem/rest'
#
#
##############################################################################
def GetMapping(base_url, ids, src_id, dst_id, fout):
  n_qry=0; n_unmapped=0; n_ambig=0; n_out=0; 
  src_info = GetSourceInfo(base_url, src_id)
  src_name = src_info['name'] if 'name' in src_info else ''
  dst_info = GetSourceInfo(base_url, dst_id)
  dst_name = dst_info['name'] if 'name' in dst_info else ''
  fout.write('%s_id\t%s_id\n'%(src_name, dst_name))
  for id_query in ids:
    n_qry+=1
    logging.info('query: "%s"'%id_query)
    rval=rest.Utils.GetURL(base_url+'/src_compound_id/%s/%d/%d'%(id_query, src_id, dst_id), parse_json=True)
    logging.debug('%s'%json.dumps(rval, sort_keys=True, indent=2))
    mols = rval
    ok=False
    hits_this=0
    for mol in mols:
      id_dst = mol['src_compound_id'] if 'src_compound_id' in mol else ''
      if id_dst:
        ok=True
        hits_this+=1
      fout.write('%s\t%s\n'%(id_query, id_dst))
      n_out+=1
    if not ok:
      n_unmapped+=1
    if hits_this>1:
      n_ambig+=1

  logging.info('n_qry: %d'%(n_qry))
  logging.info('n_unmapped: %d'%(n_unmapped))
  logging.info('n_ambig: %d'%(n_ambig))
  logging.info('n_out: %d'%(n_out))
  
##############################################################################
def GetStructure(base_url, ids, src_id, fout):
  n_qry=0; n_unmapped=0; n_ambig=0; n_out=0; 
  src_info = GetSourceInfo(base_url, src_id)
  src_name = src_info['name'] if 'name' in src_info else ''
  mol_tags=None;
  for id_query in ids:
    n_qry+=1
    logging.info('query: "%s"'%id_query)
    rval = rest.Utils.GetURL(base_url+'/structure/%s/%d'%(id_query, src_id), parse_json=True)
    logging.debug('%s'%json.dumps(rval, sort_keys=True, indent=2))
    mols = rval
    ok=False
    hits_this=0
    for mol in mols:
      if n_out==0 or not mol_tags:
        mol_tags = mol.keys()
        fout.write('\t'.join(['%s_id'%src_name]+mol_tags)+'\n')
      vals = [id_query]
      for tag in mol_tags:
        vals.append(mol[tag] if tag in mol else '')
      fout.write('\t'.join(vals)+'\n')
      n_out+=1

  logging.info('n_qry: %d'%(n_qry))
  logging.info('n_out: %d'%(n_out))

##############################################################################
def GetByInchikeys(base_url, inchikeys, fout):
  n_qry=0; n_out=0; tags=[];
  fout.write('inchikey\tchembl_id\n')
  for inchikey in inchikeys:
    n_qry+=1
    rval=None; chemb_id='';
    inchikey = re.sub(r'^InChIKey=', '', inchikey)
    logging.info('query: "%s"'%inchikey)
    rval = rest.Utils.GetURL(base_url+'/inchikey/%s'%inchikey, parse_json=True)
    chembl_id = None
    if rval:
      logging.debug('%s'%json.dumps(rval, sort_keys=True, indent=2))
      for s in rval:
        if 'src_id' in s and s['src_id']=='1':
          chembl_id = s['src_compound_id']
    if chembl_id:
      fout.write('%s\t%s\n'%(inchikey, chembl_id))
      n_out+=1

  logging.info('n_qry: %d'%(n_qry))
  logging.info('n_out: %d'%(n_out))

#############################################################################
def ListSources(base_url):
  rval = rest.Utils.GetURL(base_url+'/src_ids', parse_json=True)
  src_ids=[]
  for s in rval:
    if 'src_id' in s :
      src_ids.append(int(s['src_id']))
  for src_id in src_ids:
    info = GetSourceInfo(base_url, src_id)
    name = info['name'] if 'name' in info else ''
    name_label = info['name_label'] if 'name_label' in info else ''
    name_long = info['name_long'] if 'name_long' in info else ''
    logging.info('%2d: %s (%s)'%(src_id, name_label, name_long))
    for key in sorted(info.keys()):
      logging.debug('\t%14s: %s'%(key, info[key]))
  logging.info('source count: %d'%len(src_ids))

##############################################################################
def GetSourceInfo(base_url, src_id):
  rval=rest.Utils.GetURL(base_url+'/sources/%s'%src_id, parse_json=True)
  logging.debug(json.dumps(rval, sort_keys=True, indent=2))
  info = rval[0] if (rval and len(rval)==1) else {}
  return info

##############################################################################
if __name__=='__main__':
  epilog='''\
operations:
list_sources: all sources,
get_mapping: get mapping from src_id to dst_id,
get_by_inchikey: input IDs are InChIKeys,
get_structure: get Inchi, InchiKey from src_id,
'''
  parser = argparse.ArgumentParser(description='UniChem REST API client', epilog=epilog)
  ops=['list_sources', 'get_mapping', 'get_by_inchikey', 'get_structure']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--i", dest="ifile", help="input file of IDs")
  parser.add_argument("--ids", help="ID list (comma-separated)")
  parser.add_argument("--o", dest="ofile", help="output file (TSV)")
  parser.add_argument("--src_id", help="from-source ID code")
  parser.add_argument("--dst_id", help="to-source ID code")
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")

  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  BASE_URL='https://'+API_HOST+API_BASE_PATH

  if args.ofile:
    fout = open(args.ofile, "w")
  else:
    fout = sys.stdout

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    logging.info('input IDs: %d'%(len(ids)))
    fin.close()
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids.strip())

  if args.op == "list_sources":
    ListSources(BASE_URL)

  elif args.op == "get_mapping":
    if not ids: parser.error('--i or --ids required.')
    if not (src_id and dst_id): parser.error('--src_id and --dst_id required.')
    GetMapping(BASE_URL, ids, src_id, dst_id, fout)

  elif args.op == "get_structure":
    if not ids: parser.error('--i or --ids required.')
    if not src_id: parser.error('--src_id required.')
    GetStructure(BASE_URL, ids, src_id, fout)

  elif args.op == "get_by_inchikey":
    if not ids: parser.error('--i or --ids required.')
    GetByInchikeys(BASE_URL, ids, fout)

  else:
    parser.error('Invalid operation: %s'%(args.op))



#!/usr/bin/env python3
"""
Pharos  GraphQL API client

Depends on package python-graphql-client.
"""
###
import sys,os,json,re,time,logging
#
from python_graphql_client import GraphqlClient
#
# curl 'https://pharos-api.ncats.io/graphql' -H 'Accept-Encoding: gzip, deflate, br' -H 'Content-Type: application/json' -H 'Accept: application/json' -H 'Connection: keep-alive' -H 'DNT: 1' -H 'Origin: https://pharos-api.ncats.io' --data-binary '{"query":"query targetDetails {\n  target(q: { sym: \"ACE2\" }) {\n    name\n    tdl\n    fam\n    sym\n    description\n    novelty\n  }\n}\n"}' --compressed
#
API_ENDPOINT="https://pharos-api.ncats.io/graphql"
#
API_HEADERS = { 'Accept-Encoding': 'gzip, deflate, br',
	'Content-Type': 'application/json', 'Accept': 'application/json',
	'Connection': 'keep-alive', 'DNT': '1',
	'Origin': 'https://pharos-api.ncats.io' }
#
IDTYPES_TARGET = {
	'tcrdid':{'type':'Int'},
	'uniprot':{'type':'String'},
	'sym':{'type':'String'},
	}
IDTYPES_DISEASE = {
	'cui':{'type':'String'},
	'doid':{'type':'String'},
	'name':{'type':'String'},
	}
#
#############################################################################
def GetTargets(ids, idtype, ep=API_ENDPOINT, fout=None):
  tags=[]; n_out=0;
  client = GraphqlClient(endpoint=ep, verify=True)
  query = f"""\
query targetDetails($id: {IDTYPES_TARGET[idtype]['type']}) {{
  target(q: {{ {idtype}: $id }}) {{
    name
    tdl
    fam
    sym
    tcrdid
    uniprot
    description
    novelty
  }}
}}
"""

  for id_this in ids:
    logging.debug(f"id_this: {id_this}")
    variables = { "id": int(id_this) if IDTYPES_TARGET[idtype]['type']=='Int' else id_this }
    try:
      data = client.execute(query=query, variables=variables, headers=API_HEADERS)
    except Exception as e:
      logging.error(e)
      continue
    print(json.dumps(data, indent=2))
    n_out+=1
  logging.info(f"n_in: {len(ids)}; n_out: {n_out}")

#############################################################################
def GetDiseases(ids, idtype, ep=API_ENDPOINT, fout=None):
  tags=[]; n_out=0;
  client = GraphqlClient(endpoint=ep, verify=True)
  query = f"""\
query diseaseDetails($id: String) {{
  disease(name: $id) {{
    name
    dids {{ id, doName, dataSources, doDefinition }}
    doDescription
    uniprotDescription
    children {{
      name
      dids {{ id, doName, dataSources, doDefinition }}
      doDescription
    }}
    targetCounts {{
      name
      value
    }}
  }}
}}
"""

  for id_this in ids:
    logging.debug(f"id_this: {id_this}")
    variables = { "id": id_this }
    try:
      data = client.execute(query=query, variables=variables, headers=API_HEADERS)
    except Exception as e:
      logging.error(e)
      continue
    print(json.dumps(data, indent=2))
    n_out+=1
  logging.info(f"n_in: {len(ids)}; n_out: {n_out}")

#############################################################################
def Test(ep=API_ENDPOINT, fout=None):
  client = GraphqlClient(endpoint=ep, verify=True)
  query = """\
query diseaseDetails {{
  disease {{
    name
    dids {{
      id(id:"DOID:2841"), doName
    }}
    doDescription
    uniprotDescription
"""
  try:
    data = client.execute(query=query, headers=API_HEADERS)
  except Exception as e:
    logging.error(e)
    return
  print(json.dumps(data, indent=2))




#!/usr/bin/env python3
"""
GeneOntology client.
"""
import os,sys,re,json,time,logging
import urllib.parse
import pandas as pd

from ..util import rest

API_HOST='api.geneontology.org'
API_BASE_PATH='/api'
BASE_URL = 'https://'+API_HOST+API_BASE_PATH
#
##############################################################################
def GetEntities(ids, base_url=BASE_URL, fout=None): 
  """For only one type of entity per call (gene, term)."""
  tags=[]; df=pd.DataFrame();
  for id_this in ids:
    ent = rest.GetURL(base_url+'/bioentity/'+urllib.parse.quote(id_this), parse_json=True)
    logging.debug(json.dumps(ent, sort_keys=True, indent=2))
    if not tags: tags = ent.keys()
    df = pd.concat([df, pd.DataFrame({tags[j]:[ent[tags[j]]] for j in range(len(tags))})])
  logging.info('n_ent: {}'.format(df.shape[0]))
  if fout: df.to_csv(fout, "\t", index=False)
  return df

##############################################################################
def GetGeneTerms(ids, base_url=BASE_URL, fout=None): 
  tags=[]; df=pd.DataFrame();
  for id_this in ids:
    rval = rest.GetURL(base_url+'/bioentity/gene/{}/function'.format(urllib.parse.quote(id_this)), parse_json=True)
    assns = rval['associations'] if 'associations' in rval else []
    for assn in assns:
      logging.debug(json.dumps(assn, sort_keys=True, indent=2))
      if not tags: tags = assn.keys()
      df = pd.concat([df, pd.DataFrame({tags[j]:[assn[tags[j]]] for j in range(len(tags))})])
  logging.info('n_gene: {}; n_assn: {}'.format(len(ids), df.shape[0]))
  if fout: df.to_csv(fout, "\t", index=False)
  return df

##############################################################################



#!/usr/bin/env python3
"""
http://data.bioontology.org/documentation
The National Center for Biomedical Ontology was founded as one of the National Centers for Biomedical Computing, supported by the NHGRI, the NHLBI, and the NIH Common Fund.

An API Key is required to access any API call. It can be provided in three ways:

1. Using the apikey query string parameter.
2. Providing an Authorization header: Authorization: apikey token=your_apikey (replace `your_apikey` with your actual key)
3. When using a web browser to explore the API, if you provide your API Key once using method 1, it will be stored in a cookie for subsequent requests. You can override this by providing a different API Key in a new call.
"""
###
import sys,os,re,json,requests,urllib.parse,logging,time
import pandas as pd
#
from ..util import rest
#
API_HOST="data.bioontology.org"
API_BASE_PATH=""
#
#############################################################################
def RecommendOntologies(base_url, api_key, texts, fout):
  """This API call designed for text, not necessarily single terms."""
  #input_type={1|2} // default = 1. 1 means that the input type is text. 2 means that the input type is a list of comma separated keywords.
  tags=[]; df=pd.DataFrame(); n_err=0;
  resultTags=["coverageResult", "specializationResult", "acceptanceResult", "detailResult"];
  headers = {"Authorization":f"apikey token={api_key}"}
  for text in texts:
    url_this = base_url+f"/recommender?input={urllib.parse.quote(text)}"
    url_this += "&input_type=2"
    url_this += "&display_context=false&display_links=false"
    logging.debug(url_this)
    rval = requests.get(url_this, headers=headers)
    if not rval.ok:
      logging.error(f'{rval.status_code} : "{text}"')
      n_err+=1
      continue
    results = rval.json()
    for result in results:
      logging.debug(json.dumps(result, indent=2))
      if not tags:
        tags = list(result.keys())
        df_this = pd.DataFrame({tags[j]:([str(result[tags[j]])] if tags[j] in result else ['']) for j in range(len(tags))})
        df = pd.concat([df, df_this])
  if fout: df.to_csv(fout, "\t", index=False)
  logging.info(f"n_texts: {len(texts)}; n_out: {df.shape[0]}; n_err: {n_err}")
  return df

#############################################################################



#!/usr/bin/env python3
##############################################################################
### Utilities for Wikipathways REST API.
### See: http://www.wikipathways.org/index.php/Help:WikiPathways_Webservice/API
##############################################################################
import sys,os,re,time,logging
import urllib.parse
#
from ..util import rest
#
##############################################################################
def ListOrganisms(base_url, fout):
  n_all=0; n_out=0; n_err=0;
  url=base_url+'/listOrganisms?format=json'
  rval=rest.GetURL(url, parse_json=True)
  organisms = rval['organisms']
  fout.write('Organism\n')
  for organism in organisms:
    fout.write('%s\n'%(organism))
    n_out+=1
  logging.info('n_out: %d'%(n_out))

##############################################################################
def ListPathways(base_url, params, fout):
  n_all=0; n_out=0; n_err=0;
  url = base_url+'/listPathways?format=json'
  if params['human']:
    url+=('&organism=%s'%urllib.parse.quote('Homo sapiens'))
  rval=rest.GetURL(url, parse_json=True)
  pathways = rval['pathways']
  tags=[];
  for pathway in pathways:
    n_all+=1
    if n_all==1 or not tags:
      tags = sorted(pathway.keys())
      fout.write('\t'.join(tags)+'\n')
    vals = [pathway[tag] if tag in pathway else '' for tag in tags]
    fout.write((','.join(vals))+'\n')
    n_out+=1
  logging.info('n_all: %d; n_out: %d; n_err: %d'%(n_all,n_out,n_err))

##############################################################################
def GetPathway(base_url, id_query, ofmt, fout):
  n_all=0; n_out=0; n_err=0;
  if ofmt.lower() == 'gpml':
    url = base_url+'/getPathway?pwId=%s&revision=0'%id_query
  else:
    url = base_url+'/index.php?method=getPathwayAs&fileType=%s&pwId=%s&revision=0'%(ofmt.lower(), id_query)
  rval = rest.GetURL(url, parse_json=False)
  fout.write(rval)




#!/usr/bin/env python3
"""
PantherDb  REST API client
http://pantherdb.org/help/PANTHERhelp.jsp
"""
###
import sys,os,argparse,re,urllib.parse,logging
#
from ..util import rest
#
#############################################################################
def ListSpecies(base_url, fout):
  """
E.g.
Long Name	Short Name	NCBI Taxon Id
Homo sapiens	HUMAN	9606
  """
  n_species=0;
  url = base_url+'?type=organism'
  rval = rest.GetURL(url)
  if not rval:
    return
  for line in rval.splitlines():
    if not line: continue
    vals = re.split('\t', line)
    fout.write('\t'.join(vals)+'\n')
    n_species+=1
  logging.info('n_species = %d'%(n_species))

#############################################################################
### The following parameters are required.
### type - This refers to the search type and should be specified as "matchingOrtholog"
### inputOrganism - The organism/genome being queried by the search term(s)
### targetOrganism -  Target organism for ortholgos from search.  Multiple organisms can be specified, separated by commas
### orthologType - LDO or all - LDO will return only least diverged ortholog for each gene (single "best" ortholog), and all will return all orthologs if more than one
### searchTerm - query terms, these should optimally be Uniprot of MOD gene identifiers, but, other identifiers are supported such as gene symbols.  Maximum of 10 query terms can be submitted, separated by commas.
### For the inputOrganism and targetOrganism, the 5-letter Uniprot code is used, see the Short Name field on the summaryStats page for a list of available organisms and the associated codes.
### Example - http://www.pantherdb.org/webservices/ortholog.jsp?type=matchingOrtholog&inputOrganism=MOUSE&targetOrganism=HUMAN&orthologType=all&searchTerm=101816
### For each match, the following data are returned:
### If search term has no match in the input organism:
### <searchTerm>    "Search term not found in input organism <inputOrganism"
### If gene is found, but it has no ortholog in the database or orthologType specified is not found in the database:
### <searchTerm>     <matchedGene>    "No ortholog found in target organism <targetOrganism>"
#############################################################################
# SearchOrthologs(base_url, iorg, torg, ologtype, uniprot, fout, verbose):

#############################################################################
# Example /search.jsp?keyword=KEYWORD&type=getList&listType=LISTTYPE
# where KEYWORD is the search term and LISTTYPE = (gene|family|pathway|category)
def SearchFamilies(base_url, term, fout):
  url = base_url+'?type=getList&listType=family&keyword=%s'%(urllib.parse.quote(term))
  rval = rest.GetURL(url)
  if not rval:
    return
  logging.debug(rval)

#############################################################################
if __name__=='__main__':
  API_HOST="www.pantherdb.org"
  API_BASE_PATH="/webservices/garuda/search.jsp"
  parser = argparse.ArgumentParser(description='PantherDb REST API client utility')
  ops = ['list_species', 'search_families']
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--i", dest="ifile", help="")
  parser.add_argument("--nmax", help="list: max to return")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--query", help="search query term")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("-v", "--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url='http://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  if args.op=='list_species':
    ListSpecies(base_url, fout)

  elif args.op=='search_families':
    SearchFamilies(base_url, args.query, fout)

  else:
    parser.print_help()



#!/usr/bin/env python3
"""
TIN-X (Target Importance and Novelty Explorer) REST API client
https://www.newdrugtargets.org/
https://api.newdrugtargets.org/docs
https://api.newdrugtargets.org/targets/
"""
###
import sys,os,re,argparse,time,logging
#
from ...idg import tinx
#
##############################################################################
if __name__=='__main__':
  epilog="""\
Example IDs:
5391 (disease); DOID:9297 (DOID);
12203 (target); Q9H4B4 (UniProt);
--disease_ids needed ONLY if BOTH target and disease IDs specified, such
as for get_disease_target_articles.
"""
  parser = argparse.ArgumentParser(description="TIN-X (Target Importance and Novelty Explorer) REST API client)", epilog=epilog)
  ops = [
	"list_diseases", "list_targets", "list_articles", "list_dto",
	"get_disease", "get_disease_by_doid", "get_disease_targets", "get_disease_target_articles", "get_target", "get_target_by_uniprot", "get_target_diseases", "get_dto",
	"search_diseases", "search_targets", "search_articles", "search_dtos"
	]
  parser.add_argument("op", choices=ops, help="operation")
  parser.add_argument("--i", dest="ifile", help="input IDs or search terms")
  parser.add_argument("--ids", help="IDs (comma-separated)")
  parser.add_argument("--disease_ids", help="disease IDs (comma-separated), needed ONLY if BOTH target and disease IDs specified")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--query", help="search query")
  parser.add_argument("--api_host", default=tinx.Utils.API_HOST)
  parser.add_argument("--api_base_path", default=tinx.Utils.API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  fout = open(args.ofile, "w") if args.ofile else sys.stdout

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
  elif args.ids:
    ids = re.split('[,\s]+', args.ids.strip())
  if ids: logging.info('Input IDs: %d'%(len(ids)))
  disease_ids = re.split('[,\s]+', args.disease_ids.strip()) if args.disease_ids else None

  t0=time.time()

  base_url = 'https://'+args.api_host+args.api_base_path

  if args.op=='list_targets':
    tinx.Utils.ListTargets(base_url, fout)

  elif args.op=='list_diseases':
    tinx.Utils.ListDiseases(base_url, fout)

  elif args.op=='list_articles':
    tinx.Utils.ListArticles(base_url, fout)

  elif args.op=='list_dto':
    tinx.Utils.ListDTO(base_url, fout)

  elif args.op=='get_disease':
    tinx.Utils.GetDisease(ids, base_url, fout)

  elif args.op=='get_disease_by_doid':
    tinx.Utils.GetDiseaseByDOId(ids, base_url, fout)

  elif args.op=='get_disease_targets':
    if not ids: parser.error(f"--i or --ids required for {args.op}")
    tinx.Utils.GetDiseaseTargets(ids, base_url, fout)

  elif args.op=='get_disease_target_articles':
    if not disease_ids: parser.error(f"--disease_ids required for {args.op}")
    tinx.Utils.GetDiseaseTargetArticles(disease_ids, ids, base_url, fout)

  elif args.op=='get_target':
    tinx.Utils.GetTarget(ids, base_url, fout)

  elif args.op=='get_target_by_uniprot':
    tinx.Utils.GetTargetByUniprot(ids, base_url, fout)

  elif args.op=='get_target_diseases':
    tinx.Utils.GetTargetDiseases(ids, base_url, fout)

  elif args.op=='search_diseases':
    if not args.query: parser.error(f"--query required for {args.op}")
    tinx.Utils.SearchDiseases(args.query, base_url, fout)

  elif args.op=='search_targets':
    if not args.query: parser.error(f"--query required for {args.op}")
    tinx.Utils.SearchTargets(args.query, base_url, fout)

  elif args.op=='search_articles':
    if not args.query: parser.error(f"--query required for {args.op}")
    tinx.Utils.SearchArticles(args.query, base_url, fout)

  else:
    parser.error(f'Unknown operation: {args.op}')

  logging.info(('elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0)))))



#!/usr/bin/env python3
#############################################################################
### BindingDb Utilities
### http://www.bindingdb.org/bind/BindingDBRESTfulAPI.jsp
### http://www.bindingdb.org/axis2/services/BDBService/getLigandsByUniprots?uniprot=P35355,Q8HZR1&cutoff=1000&code=0&response=application/json
#############################################################################
import sys,os,re,time,json,logging
import urllib.parse

from ..util import rest
#
##############################################################################
def GetLigandsByUniprot(base_url, ids, ic50_max, fout):
  n_out=0; tags=None;
  for id_this in ids:
    rval = rest.GetURL(base_url+'/getLigandsByUniprots?uniprot=%s&cutoff=%d&response=application/json'%(id_this, ic50_max), parse_json=True)
    logging.debug(json.dumps(rval, sort_keys=True, indent=2))
    ligands = rval["getLigandsByUniprotsResponse"]["affinities"] if "getLigandsByUniprotsResponse" in rval and "affinities" in rval["getLigandsByUniprotsResponse"] else []
    for ligand in ligands:
      if not tags:
        tags = ligand.keys()
        fout.write("\t".join(tags)+"\n")
      vals = [(str(ligand[tag]) if tag in ligand else '') for tag in tags]
      fout.write("\t".join(vals)+"\n")
  logging.info("n_out: {}".format(n_out))

##############################################################################
def GetTargetsByCompound(base_url, smiles, sim_min, fout):
  rval = rest.GetURL(base_url+'/getTargetByCompound?smiles=%s&cutoff=%.2f'%(urllib.parse.quote(smiles),
sim_min), parse_xml=True)
  fout.write(rval.tostring())

##############################################################################



#!/usr/bin/env python
#############################################################################
### BindingDb Utilities
### http://www.bindingdb.org/bind/BindingDBRESTfulAPI.jsp
### XML output only.
#############################################################################
import sys,os,re,argparse,time,logging
#
from .. import bindingdb
#
##############################################################################
if __name__=='__main__':
  epilog = "Example Uniprot IDs: P35355, Q8HZR1"
  API_HOST='www.bindingdb.org'
  API_BASE_PATH='/axis2/services/BDBService'
  ops = ["get_ligands_by_uniprot", "get_targets_by_compound"]
  parser = argparse.ArgumentParser( description='BindingDb REST API client', epilog=epilog)
  parser.add_argument("op", choices=ops, help='OPERATION (select one)')
  parser.add_argument("--i", dest="ifile", help="input file, Uniprot IDs")
  parser.add_argument("--ids", help="Uniprot IDs")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--smiles", help="compound query")
  parser.add_argument("--ic50_max", type=int, default=100)
  parser.add_argument("--sim_min", type=float, default=0.85)
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("-v", "--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  api_base_url='http://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  t0=time.time()

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
  elif args.ids:
    ids = re.split('[, ]+', args.ids.strip())
  if len(ids)>0: logging.info('Input IDs: %d'%(len(ids)))

  if args.op=="get_ligands_by_uniprot":
    bindingdb.GetLigandsByUniprot(api_base_url, ids, args.ic50_max, fout)

  elif args.op=="get_targets_by_compound":
    bindingdb.GetTargetsByCompound(api_base_url, args.smiles, args.sim_min, fout)

  else:
    parser.error('Operation invalid: {}'.format(args.op))

  logging.info('Elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0))))




#!/usr/bin/env python3
"""
Utility functions for the PubChem PUG REST API.

* https://pubchemdocs.ncbi.nlm.nih.gov/pug-rest

	<DOMAIN>/<NAMESPACE>/<IDENTIFIERS>

	compound/{cid|name|smiles|inchi|inchikey|listkey}
	compound/substructure/{smiles|inchi}
	substance/{sid|name|listkey}
	substance/sourceid/<source name> 
	substance/sourceall/<source name> 
	substance/sid/<SID>/cids/JSON
	substance/sid/<SID>/cids/XML?cids_type=all
	assay/{aid|listkey} 
 	assay/type/{all|confirmatory|doseresponse|onhold|panel|rnai|screening|summary}
 	assay/sourceall/<source name>
	sources/{substance|assay}

  <IDENTIFIERS> = comma-separated list of positive integers (cid, sid, aid) or
  strings (source, inchikey, listkey); single identifier string
  (name, smiles; inchi by POST only)
"""
###
import sys,os,io,re,csv,json,pandas,math,time,logging,tempfile,tqdm,tqdm.auto
from xml.etree import ElementTree
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
import urllib.request,urllib.parse
import pandas as pd
#
OUTCOME_CODES = {
        'inactive':1,
        'active':2,
        'inconclusive':3,
        'unspecified':4,
        'probe':5}
#
API_HOST='pubchem.ncbi.nlm.nih.gov'
API_BASE_PATH='/rest/pug'
BASE_URL="https://"+API_HOST+API_BASE_PATH
#
#############################################################################
def OutcomeCode(txt):
  return OUTCOME_CODES[txt.lower()] if txt.lower() in OUTCOME_CODES else OUTCOME_CODES['unspecified']

#############################################################################
def ListSources(src_type, base_url=BASE_URL, fout=None):
  rval = requests.get(base_url+f"/sources/{src_type}/JSON").json()
  logging.debug(json.dumps(rval,indent=2))
  sources = rval['InformationList']['SourceName'] if 'InformationList' in rval and 'SourceName' in rval['InformationList'] else []
  n_src=0;
  for source in sorted(sources):
    fout.write(f"{source}\n")
    n_src+=1
  logging.info(f"n_src ({src_type}): {n_src}")

##############################################################################
def GetSID2SDF(id_query, base_url=BASE_URL):
  url = (base_url+f"/substance/sid/{id_query}/SDF")
  txt = requests.get(url).text
  return txt

#############################################################################
def GetSID2CID(sids, base_url=BASE_URL, fout=None):
  cids=set(); tq=None; df=None;
  if fout: fout.write("SID\tCID\n")
  for sid in sids:
    rval = requests.get(base_url+f"/substance/sid/{sid}/cids/JSON?cids_type=standardized").json()
    infos = rval['InformationList']['Information'] if 'InformationList' in rval and 'Information' in rval['InformationList'] else []
    for info in infos:
      cids_this = info['CID'] if 'CID' in info else []
      for cid in cids_this:
        cids.add(str(cid))
        if fout: fout.write(f"{sid}\t{cid}\n")
  return list(cids)

#############################################################################
def GetSID2Smiles(sids, base_url=BASE_URL, fout=None):
  cids = GetSID2CID(sids, base_url)
  GetCID2Smiles(cids, base_url, fout)

#############################################################################
def GetCID2SID(cids, base_url=BASE_URL, fout=None):
  sids=set(); df=None;
  if fout: fout.write("CID\tSID\n")
  for cid in cids:
    time.sleep(0.01) #Kludge fix: requests.exceptions.ConnectionError: ('Connection aborted.', ConnectionResetError(104, 'Connection reset by peer'))
    url_this = f"{base_url}/compound/cid/{cid}/sids/JSON"
    response = requests.get(url_this)
    if response.status_code!=200:
      logging.debug(f"status_code: {response.status_code}; {url_this}")
      continue
    try:
      rval = response.json()
    except Exception as e:
      logging.error(f"{str(e)}")
      continue
    infos = rval['InformationList']['Information'] if 'InformationList' in rval and 'Information' in rval['InformationList'] else []
    for info in infos:
      sids_this = info['SID'] if 'SID' in info else []
      for sid in sids_this:
        sids.add(str(sid))
        if fout: fout.write(f"{cid}\t{sid}\n")
  return list(sids)

#############################################################################
### Must quote forward slash '/' in smiles.  2nd param in quote()
### specifies "safe" chars to not quote, '/' being the default, so specify
### '' for no safe chars.
#############################################################################
def GetSmiles2CID(smis, base_url=BASE_URL, fout=None):
  n_out=0; tq=None; df=None;
  #fout.write("CID\tSMILES\tName\n")
  for i_smi in tqdm.auto.trange(len(smis)):
    smi = smis[i_smi]
    name = re.sub(r'^[\S]+\s', '', smi) if re.search(r'^[\S]+\s', smi) else ""
    smi = re.sub(r'\s.*$', '', smi)
    rval = requests.get(base_url+f"/compound/smiles/{urllib.parse.quote(smi, '')}/cids/JSON").json()
    if not rval: continue
    cids = rval['IdentifierList']['CID'] if 'IdentifierList' in rval and 'CID' in rval['IdentifierList'] else []
    #for cid in cids:
    #  fout.write(f"{cid}\t{smi}\t{name}\n")
    #  n_out+=1
    df_this = pd.DataFrame({"CID":cids, "SMILES":smi, "Name":name})
    if fout is not None: df_this.to_csv(fout, "\t", header=bool(n_out==0), index=False)
    else: df = pd.concat([df, df_this])
    n_out+=len(cids)
  logging.info(f"SMIs: {len(smis)}; CIDs out: {n_out}")
  return df

#############################################################################
def GetCID2AssaySummary(ids, base_url=BASE_URL, fout=None):
  """Example CIDs: 2519 (caffeine), 3034034 (quinine)"""
  n_out=0; df=None;
  for i_cid in tqdm.auto.trange(len(ids), desc="CIDs"):
    id_this = ids[i_cid]
    rval = requests.get(base_url+f"/compound/cid/{id_this}/assaysummary/CSV").text
    if not rval: continue
    df_this = pandas.read_csv(io.StringIO(rval), sep=',')
    if fout is not None: df_this.to_csv(fout, sep='\t', index=False, header=bool(n_out==0))
    else: df = pd.concat([df, df_this])
    n_out += df_this.shape[0]
  logging.info(f"CIDs: {len(ids)}; assay summaries out: {n_out}")
  return df

#############################################################################
def GetSID2AssaySummary(ids, base_url=BASE_URL, fout=None):
  n_out=0; df=None;
  for i_sid in tqdm.auto.trange(len(ids), desc="SIDs"):
    id_this = ids[i_sid]
    rval = requests.get(base_url+f"/substance/sid/{id_this}/assaysummary/CSV").text
    if not rval: continue
    df_this = pandas.read_csv(io.StringIO(rval), sep=',')
    if fout is not None: df_this.to_csv(fout, sep='\t', index=False, header=bool(n_out==0))
    else: df = pd.concat([df, df_this])
    n_out += df_this.shape[0]
  logging.info(f"SIDs: {len(ids)}; assay summaries out: {n_out}")
  return df

#############################################################################
def GetCID2Inchi(ids, base_url=BASE_URL, fout=None):
  PROPTAGS = [ "InChIKey", "InChI" ]
  ids_dict = {'cid':(','.join(map(lambda x:str(x), ids)))}
  url = (base_url+f"/compound/cid/property/{','.join(PROPTAGS)}/CSV")
  response = requests.post(url, headers={'Accept':'text/CSV', 'Content-type':'application/x-www-form-urlencoded'}, data=ids_dict)
  df = pandas.read_csv(io.StringIO(response.text), sep=',')
  if fout is not None: df.to_csv(fout, sep='\t', index=False)
  logging.info(f"Input IDs: {len(ids)}; Output InChIs: {df.shape[0]}")
  return df

##############################################################################
def GetCID2SDF(ids, base_url=BASE_URL, fout=None):
  """Request in chunks.  Works for 50, and not for 200 (seems to be a limit)."""
  nchunk=50; nskip_this=0; n_out=0; tq=None; txt_out="";
  while True:
    if tq is None: tq = tqdm.tqdm(total=len(ids), unit="cids")
    if nskip_this>=len(ids): break
    ids_this = ids[nskip_this:nskip_this+nchunk]
    idstr = (','.join(map(lambda x:str(x), ids_this)))
    #rval = rest.Utils.PostURL(base_url+'/compound/cid/SDF', data={'cid':idstr})
    response = requests.post(base_url+'/compound/cid/SDF', data={'cid':idstr})
    if fout is not None: fout.write(response.text)
    else: txt_out += response.text
    n_out += len(re.findall(r'^\$\$\$\$$', response.text, re.M))
    nskip_this+=nchunk
    tq.update(n=len(ids_this))
  tq.close()
  logging.info(f"SDFs out: {n_out}")
  return txt_out

#############################################################################
def GetSID2SDF(ids, skip, nmax, base_url=BASE_URL, fout=None):
  """Faster via POST(?). Request in chunks.  Works for 50, and not
  for 200 (seems to be a limit)."""
  n_out=0; tq=None; txt_out="";
  if skip: logging.debug(f"skip: [1-{skip}]")
  nchunk=50; nskip_this=skip;
  while True:
    if tq is None: tq = tqdm.tqdm(total=len(ids), unit="cids")
    if nskip_this>=len(ids): break
    nchunk = min(nchunk, nmax-(nskip_this-skip))
    n_sid_in+=nchunk
    ids_this = ids[nskip_this:nskip_this+nchunk]
    idstr = (','.join(map(lambda x:str(x), ids_this)))
    #rval = rest.Utils.PostURL(base_url+'/substance/sid/SDF', data={'sid':idstr})
    response = requests.post(base_url+'/substance/sid/SDF', data={'sid':idstr})
    if fout is not None: fout.write(response.text)
    else: txt_out += response.text
    n_out += len(re.findall(r'^\$\$\$\$$', response.text, re.M))
    nskip_this+=nchunk
    if nmax and (nskip_this-skip)>=nmax:
      logging.info(f"NMAX limit reached: {nmax}")
      break
    tq.update(n=len(ids_this))
  tq.close()
  logging.info(f"SIDs: {len(sids)}; SDFs out: {n_out}")
  return txt_out

#############################################################################
def GetCID2Smiles(ids, base_url=BASE_URL, fout=None):
  """Returns Canonical and Isomeric SMILES."""
  PROPTAGS = ['CanonicalSMILES', 'IsomericSMILES']
  url = (base_url+"/compound/cid/property/{}/CSV".format(','.join(PROPTAGS)))
  nchunk=50; nskip_this=0; df=None; tq=None;
  n_in=0; n_out=0; n_err=0; results=[];
  while True:
    if tq is None: tq = tqdm.tqdm(total=len(ids), unit="mols")
    if nskip_this>=len(ids): break
    ids_this = ids[nskip_this:nskip_this+nchunk]
    n_in+=len(ids_this)
    idstr = (','.join(map(lambda x:str(x), ids_this)))
    response = requests.post(url, data={'cid':idstr})
    df_this = pandas.read_csv(io.StringIO(response.text), sep=',')
    if fout is not None:
      df_this.to_csv(fout, sep='\t', index=False, header=bool(n_out==0))
    else:
      df = pd.concat([df, df_this])
    nskip_this+=nchunk
    n_out+=len(ids_this)
    tq.update(n=len(ids_this))
  tq.close()
  logging.info(f"Input IDs: {len(ids)}; Output records: {n_out}")
  return df

#############################################################################
def GetCID2Properties(ids, base_url=BASE_URL, fout=None):
  PROPTAGS = ["CanonicalSMILES", "IsomericSMILES", "InChIKey", "InChI", "MolecularFormula", "HeavyAtomCount", "MolecularWeight", "XLogP", "TPSA"]
  url = (base_url+"/compound/cid/property/{}/CSV".format(','.join(PROPTAGS)))
  nchunk=50; nskip_this=0; df=None; tq=None;
  n_in=0; n_out=0; n_err=0; results=[];
  while True:
    if tq is None: tq = tqdm.tqdm(total=len(ids), unit="mols")
    if nskip_this>=len(ids): break
    ids_this = ids[nskip_this:nskip_this+nchunk]
    n_in+=len(ids_this)
    idstr = (','.join(map(lambda x:str(x), ids_this)))
    response = requests.post(url, headers={'Accept':'text/CSV', 'Content-type':'application/x-www-form-urlencoded'}, data={'cid':idstr})
    try:
      df_this = pandas.read_csv(io.StringIO(response.text), sep=',')
    except Exception as e:
      logging.error(f"{e}")
      logging.debug(response.text)
      continue
    if fout is not None:
      df_this.to_csv(fout, sep='\t', index=False, header=bool(n_out==0))
    else:
      df = pd.concat([df, df_this])
    nskip_this+=nchunk
    n_out+=len(ids_this)
    tq.update(n=len(ids_this))
  tq.close()
  logging.info(f"Input IDs: {len(ids)}; Output records: {n_out}")
  return df

#############################################################################
def GetCID2Descriptions(ids, base_url=BASE_URL, fout=None):
  n_out=0; tags=None; df=None;
  for i in tqdm.tqdm(range(len(ids))):
    id_this = ids[i]
    response = requests.get(base_url+f"/compound/cid/{id_this}/description/JSON")
    if response.status_code!=200:
      logging.error(f"status_code: {response.status_code}")
      continue
    rval = response.json()
    infos = rval['InformationList']['Information'] if 'InformationList' in rval and 'Information' in rval['InformationList'] else []
    for info in infos:
      if 'Description' not in info: continue
      if not tags:
        tags = list(info)
      df_this = pd.DataFrame({tag:([info[tag]] if tag in info else ['']) for tag in tags})

      if fout is not None: df_this.to_csv(fout, "\t", header=bool(n_out==0), index=False)
      else: df = pd.concat([df, df_this])
      n_out+=df_this.shape[0]
  logging.info(f"Input IDs: {len(ids)}; Output records: {n_out}")
  return df

#############################################################################
def Inchi2CID(inchis, base_url=BASE_URL, fout=None):
  '''	Must be POST with "Content-Type: application/x-www-form-urlencoded"
	or "Content-Type: multipart/form-data" with the POST body formatted accordingly.
	See: http://pubchem.ncbi.nlm.nih.gov/pug_rest/PUG_REST.html and
	http://pubchem.ncbi.nlm.nih.gov/pug_rest/PUG_REST_Tutorial.html
  '''
  n_out=0; tq=None; df=None;
  url = base_url+"/compound/inchi/cids/TXT"
  for inchi in inchis:
    if tq is None: tq = tqdm.tqdm(total=len(ids), unit="inchis")
    logging.info(f"inchi='{inchi}'")
    body_dict={'inchi':urllib.parse.quote(inchi)}
    #rval = rest.Utils.PostURL(url=url, headers={'Content-Type':'application/x-www-form-urlencoded','Accept':'text/plain'}, data=body_dict)
    response = requests.post(url=url, headers={'Content-Type':'application/x-www-form-urlencoded','Accept':'text/plain'}, data=body_dict)
    lines = response.text.splitlines()
    cids_this=set() 
    for line in lines:
      if re.match(r'[\d]+$', line.strip()):
        cid = line.strip()
        cids_this.add(cid)
    df_this = pd.DataFrame({"InChI":inchi, "CID":list(cids_this)})
    if fout is not None: df_this.to_csv(fout, "\t", header=bool(n_out==0))
    else: df = pd.concat([df, df_this])
    n_out += len(cids_this)
    tq.update(n=1)
  tq.close()
  logging.info(f"n_out: {n_out}")
  return df

#############################################################################
def GetAssayName(aids, base_url=BASE_URL, fout=None):
  n_out=0; tq=None; df=None;
  for aid in aids:
    if tq is None: tq = tqdm.tqdm(total=len(ids), unit="assays")
    xmlstr = requests.get(base_url+f"/assay/aid/{aid}/description/XML").text
    name, source = AssayXML2NameAndSource(xmlstr)
    df_this = pd.DataFrame({"AID":[aid], "Name":[name], "Source":[source]})
    if fout is not None: df_this.to_csv(fout, "\t", header=bool(n_out==0))
    else: df = pd.concat([df, df_this])
    n_out+=1
    tq.update(n=1)
  tq.close()
  logging.info(f"n_out: {n_out}")
  return df

#############################################################################
def GetAssayDescriptions(ids, skip, nmax, base_url=BASE_URL, fout=None):
  """Example AIDs: 527,159014"""
  n_in=0; n_out=0; tq=None; df=None;
  for aid in ids:
    n_in+=1
    if skip and n_in<skip: continue
    if nmax and n_out==nmax: break
    if tq is None: tq = tqdm.tqdm(total=len(ids), unit="assays")
    url = (base_url+f"/assay/aid/{aid}/description/JSON")
    rval = requests.get(url).json()
    logging.debug(json.dumps(rval, indent=2))
    assays = rval['PC_AssayContainer'] if 'PC_AssayContainer' in rval else []
    for assay in assays:
      descr = assay['assay']['descr'] if 'assay' in assay and 'descr'  in assay['assay'] else None
      if not descr: continue
      #aid = str(descr['aid']['id'])
      name = descr['name']
      source = descr['aid_source']['db']['name']
      revision = str(descr['revision'])
      project_category = str(descr['project_category'])
      activity_outcome_method = str(descr['activity_outcome_method'])
      assay_group = descr['assay_group'] if 'assay_group' in descr else '' #list of PMIDs
      description_lines = descr['description'] #list of lines
      comment_lines = descr['comment'] #list of lines
      description = (' ').join(description_lines)
      comment = (' ').join(comment_lines)
      df_this = pd.DataFrame({
	"aid":[aid],
	"name":[name],
	"source":[source],
	"revision":[revision],
	"assay_group":[(';'.join(assay_group))],
	"project_category":[project_category],
	"activity_outcome_method":[activity_outcome_method],
	"description":[description],
	"comment":[comment]
	})
      if fout is not None: df_this.to_csv(fout, "\t", header=bool(n_out==0), index=False)
      else: df = pd.concat([df, df_this])
      n_out+=1
    tq.update(n=1)
  tq.close()
  logging.info(f"n_out: {n_out}")
  return df

#############################################################################
def GetAssaySIDs(aids, skip, nmax, base_url=BASE_URL, fout=None):
  if skip: logging.info(f"SKIP: [1-{skip}]")
  if nmax: logging.info(f"NMAX: {nmax}")
  n_aid_in=0; n_aid_done=0; n_out=0; nchunk_sid=100; df=None; tags=None;
  tqa=None;
  for aid in aids:
    if tqa is None: tqa = tqdm.tqdm(total=len(aids), unit="assays")
    n_aid_in+=1
    if skip and n_aid_in<skip: continue
    if nmax and n_aid_done>=nmax: break
    n_aid_done+=1
    logging.debug(f"Request: ({n_aid_done}) [AID={aid}]")
    url = base_url+f"/assay/aid/{aid}/concise/JSON"
    rval = requests.get(url).json()
    logging.debug(json.dumps(rval, indent=2))
    tags_this = rval["Table"]["Columns"]["Column"]
    logging.debug(f"tags_this = {str(tags_this)}")
    j_sid=None; j_cid=None; 
    if not tags:
      tags = tags_this
      j_sid = tags.index("SID") if "SID" in tags else None
      j_cid = tags.index("CID") if "CID" in tags else None
    rows = rval["Table"]["Row"]
    logging.info(f"[AID={aid}] results/activities: {len(rows)}")
    tqb=None;
    for row in rows:
      if tqb is None: tqb = tqdm.tqdm(total=len(rows), unit="results")
      cells = row["Cell"]
      sid = cells[j_sid]
      cid = cells[j_cid]
      df_this = pd.DataFrame({"AID":[aid], "SID":[sid], "CID":[cid]})
      df = pd.concat([df, df_this])
      tqb.update()
    tqb.close()
    tqa.update()
  tqa.close()
  logging.info(f"AIDs: {df.AID.nunique()}; SIDs: {df.SID.nunique()}; CIDs: {df.CID.nunique()}")
  df.drop_duplicates(inplace=True)
  df.sort_values(by=["AID", "SID"], inplace=True)
  if fout is not None: df.to_csv(fout, "\t", index=False)
  return df

#############################################################################
def GetAssaySIDResults(aids, sids, skip, nmax, base_url=BASE_URL, fout=None):
  '''One CSV line for each activity.  skip and nmax applied to AIDs.
In this version of the function, use the "concise" mode to download full data
for each AID, then iterate through SIDs and use local hash.
'''
  if skip: logging.info(f"skip: [1-{skip}]")
  if nmax: logging.info(f"NMAX: {nmax}")
  n_aid_in=0; n_aid_done=0; n_out=0; df=None; tqa=None;
  j_tag={};
  tags = None
  for aid in aids:
    if tqa is None: tqa = tqdm.tqdm(total=len(aids), unit="assays")
    n_aid_in+=1
    n_out_this=0;
    if skip and n_aid_in<skip: continue
    if nmax and n_aid_done>=nmax: break
    n_aid_done+=1
    logging.debug(f"Request: ({n_aid_done}) [AID={aid}] SID count: {len(sids)}")
    url = base_url+f"/assay/aid/{aid}/concise/JSON"
    rval = requests.get(url).json()
    if not tags:
      tags = rval["Table"]["Columns"]["Column"]
      logging.debug(f"tags= {str(tags)}")
      for tag in tags:
        j_tag[tag] = tags.index(tag)
    rows = rval["Table"]["Row"]
    tqb=None;
    for row in rows:
      if tqb is None: tqb = tqdm.tqdm(total=len(rows), unit="results")
      cells = row["Cell"]
      logging.debug(json.dumps(cells, indent=2))
      df_this = pd.DataFrame({tag:[cells[j_tag[tag]]] for tag in tags})
      df = pd.concat([df, df_this])
      tqb.update()
    tqb.close()
    tqa.update()
  tqa.close()
  logging.info(f"AIDs: {df.AID.nunique()}; SIDs: {df.SID.nunique()}; CIDs: {df.CID.nunique()}; Activity results: {df.shape[0]}")
  for key,val in df['Bioactivity Outcome'].value_counts().iteritems():
    logging.info(f'Bioactivity_Outcome = {key:>12}: {val:6d}')
  df.drop_duplicates(inplace=True)
  df.sort_values(by=["AID", "SID"], inplace=True)
  if fout is not None: df.to_csv(fout, "\t", index=False)
  return df

#############################################################################
def AssayXML2NameAndSource(xmlstr):
  '''Required: xpath - XPath Queries For DOM Trees, http://py-dom-xpath.googlecode.com/'''
  logging.debug(xmlstr)
  root = ElementTree.fromstring(xmlstr)
  ns = { "pc":"http://www.ncbi.nlm.nih.gov" }
  name = root.find("pc:PC-AssaySubmit/pc:PC-AssaySubmit_assay/pc:PC-AssaySubmit_assay_descr/pc:PC-AssayDescription/pc:PC-AssayDescription_name", ns).text  ##1st
  source = root.find("pc:PC-AssaySubmit/pc:PC-AssaySubmit_assay/pc:PC-AssaySubmit_assay_descr/pc:PC-AssayDescription/pc:PC-AssayDescription_aid-source/pc:PC-Source/pc:PC-Source_db/pc:PC-DBTracking/pc:PC-DBTracking_name",
ns).text
  return name,source

#############################################################################
def GetSID2Synonyms(ids, base_url=BASE_URL, fout=None):
  n_out=0; df=None;

  retry_strategy = Retry(
	total=30,
	backoff_factor=2,
	status_forcelist=[413, 429, 500, 502, 503, 504],
	method_whitelist=["HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE"]
	)
  adapter = HTTPAdapter(max_retries=retry_strategy)
  session = requests.Session()
  session.mount("https://", adapter)
  session.mount("http://", adapter)

  for i_sid in range(len(ids)):
    id_this = ids[i_sid]
    url_this = f"{base_url}/substance/sid/{id_this}/synonyms/JSON"
    response = session.get(url_this)
    if response.status_code!=200: #404 is normal
      logging.debug(f"Status code: {response.status_code}; url_this: {url_this}")
      continue
    rval = response.json()
    infos = rval['InformationList']['Information'] if 'InformationList' in rval and 'Information' in rval['InformationList'] else []
    for info in infos:
      synonyms_this = info['Synonym'] if 'Synonym' in info else []
      df_this = pd.DataFrame({"SID":id_this, "Synonym":synonyms_this})
      if fout is not None: df_this.to_csv(fout, "\t", header=bool(n_out==0), index=False)
      else: df = pd.concat([df, df_this])
      n_out+=len(synonyms_this)
  return df

#############################################################################
def GetCID2Synonyms(ids, skip, nmax, nmax_per_cid, base_url=BASE_URL, fout=None):
  sids_all = set([])
  i_cid=0; n_out=0; df=None;
  for i_cid in tqdm.auto.trange(len(ids), desc="CIDs"):
    id_this = ids[i_cid]
    if skip and i_cid<skip: continue
    sids_this = GetCID2SID([id_this], base_url)
    sids_all |= set(sids_this)
    synonyms_this_cid = set()
    for i_sid in tqdm.auto.trange(len(sids_this), desc="SIDs", leave=False):
      sid = sids_this[i_sid]
      synonyms_this_sid = GetSID2Synonyms([sid], base_url)
      if synonyms_this_sid is None: continue
      synonyms_this_sid = synonyms_this_sid.Synonym.tolist()
      synonyms_this_cid |= set(synonyms_this_sid)
    synonyms_this_cid_nice = SortCompoundNamesByNiceness(list(synonyms_this_cid))
    df_this = pd.DataFrame({"CID":id_this, "Synonym":synonyms_this_cid_nice}) 
    if df_this.shape[0]>nmax_per_cid:
      df_this = df_this[:nmax_per_cid]
      logging.debug(f"{i_cid+1}. CID={id_this}: synonyms out+truncated=all: {nmax_per_cid}+{len(synonyms_this_cid_nice)-nmax_per_cid}={len(synonyms_this_cid_nice)}")
    if fout is not None:
      df_this.to_csv(fout, "\t", header=bool(n_out==0), index=False)
      fout.flush()
    else: df = pd.concat([df, df_this])
    n_out+=len(synonyms_this_cid_nice)
    logging.debug(f"{i_cid+1}. CID={id_this}: SIDs: {len(sids_this)}; synonyms: {len(synonyms_this_cid)} ({min(len(synonyms_this_cid_nice), nmax_per_cid)})")
    if nmax and (i_cid+1)>=(skip+nmax): break
  logging.info(f"Totals: CIDs: {len(ids)}; SIDs: {len(sids_all)}; Synonyms: {n_out}")
  return df

#############################################################################
def GetCID2Nicename(ids, skip, nmax, base_url=BASE_URL, fout=None):
  return GetCID2Synonyms(ids, skip, nmax, 1, base_url, fout)

#############################################################################
def GetName2SID(names, skip, nmax, base_url=BASE_URL, fout=None):
  n_sid=0; sids_all=set(); df=None;
  for i_name in tqdm.auto.trange(len(names), leave=False):
    name = names[i_name]
    if skip and (i_name+1)<=skip: continue
    rval = requests.get(base_url+f"/substance/name/{urllib.parse.quote(name)}/sids/JSON").json()
    sids_this = rval['IdentifierList']['SID'] if 'IdentifierList' in rval and 'SID' in rval['IdentifierList'] else []
    for sid in sids_this:
      df_this = pd.DataFrame({"Name":[name], "SID":[sid]})
      if fout is not None: df_this.to_csv(fout, "\t", header=bool(n_sid==0), index=False)
      else: df = pd.concat([df, df_this])
      n_sid+=1
    sids_all |= set(sids_this)
    if nmax and (i_name+1)>=(skip+nmax): break
  logging.info(f"n_name: {len(names)}; n_sid: {n_sid}; n_sid_unique: {len(sids_all)}")
  return df

#############################################################################
def GetName2CID(names, skip, nmax, base_url=BASE_URL, fout=None):
  n_cid=0; cids_all=set(); n_sid=0; sids_all=set(); df=None;
  for i_name in tqdm.auto.trange(len(names), leave=False):
    name = names[i_name]
    if skip and (i_name+1)<=skip: continue
    #sids_this = GetName2SID([name], base_url)
    rval = requests.get(base_url+f"/substance/name/{urllib.parse.quote(name)}/sids/JSON").json()
    sids_this = rval['IdentifierList']['SID'] if 'IdentifierList' in rval and 'SID' in rval['IdentifierList'] else []
    sids_all |= set(sids_this)
    for sid in sids_this:
      n_sid+=1
      cids_this = list(set(GetSID2CID([sid], base_url)) )
      for cid in cids_this:
        df_this = pd.DataFrame({"Name":[name], "SID":[sid], "CID":[cid]})
        if fout is not None: df_this.to_csv(fout, "\t", header=bool(n_cid==0), index=False)
        else: df = pd.concat([df, df_this])
        n_cid+=1
      cids_all |= set(cids_this)
    if nmax and (i_name+1)>=(skip+nmax): break
  logging.info(f"n_name: {len(names)}; n_sid: {n_sid}; n_sid_unique: {len(sids_all)}; n_cid: {n_cid}; n_cid_unique: {len(cids_all)}")
  return df

#############################################################################
def GetName2Synonyms(names, skip, nmax, base_url=BASE_URL, fout=None):
  n_synonym=0; sids_all=set(); synonyms_all=set(); df=None;
  if fout: fout.write("Name\tSID\tSynonym\n")
  for i_name in tqdm.auto.trange(len(names), leave=False):
    name = names[i_name]
    sids_this = GetName2SID([name], 0, None, base_url)
    sids_all |= set(sids_this)
    for sid in sids_this:
      rval = requests.get(base_url+f"/substance/sid/{sid}/synonyms/JSON").json()
      infos = rval['InformationList']['Information'] if 'InformationList' in rval and 'Information' in rval['InformationList'] else []
      synonyms_this_sid = set()
      for info in infos:
        synonyms_this = info['Synonym'] if 'Synonym' in info else []
        synonyms_this_sid |= set(synonyms_this)
        for synonym in synonyms_this:
          n_synonym+=1
          if fout: fout.write(f"{name}\t{sid}\t{synonym}\n")
      synonyms_all |= set(synonyms_this_sid)
  logging.info(f"n_name: {len(names)}; n_synonym: {n_synonym}; n_synonym_unique: {len(synonyms_all)}; n_sids_unique: {len(sids_all)}")

#############################################################################
def NameNicenessScore(name, len_optimal=9):
  """Heuristic for human comprehensibility. In PubChem, a 'name' may be
any ID, often not useful."""
  score=0;
  pat_proper = re.compile(r'^[A-Z][a-z]+$')
  pat_text = re.compile(r'^[A-Za-z ]+$')
  if pat_proper.match(name): score+=100
  elif pat_text.match(name): score+=50
  elif re.match(r'^[A-z][A-z][A-z][A-z][A-z][A-z][A-z].*$', name): score+=10
  if re.search(r'[\[\]\{\}]', name): score-=50
  if re.search(r'[_/\(\)]', name): score-=10
  if re.search(r'\d', name): score-=10
  score -= math.fabs(len_optimal-len(name))
  return score

#############################################################################
def SortCompoundNamesByNiceness(names):
  names_scored = {n:NameNicenessScore(n) for n in names}
  names_ranked = [[score,name] for name,score in names_scored.items()]
  names_ranked.sort()
  names_ranked.reverse()
  return [name for score,name in names_ranked]

#############################################################################
def GetCpdAssayStats(cid, smiles, aidset, aidhash, base_url, fout_mol, fout_act):
  """
THIS FUNCTION NEEDS WORK.
From PubChem PUG REST API, determine activity stats for compounds.
 
 Input: smiles and CIDs
 Output (1):
   SMILES CID SID aTested aActive wTested wActive 
     where:
   aTested - how many assays where that cpd has been tested
   aActive - how many assays where that cpd has been tested active
   wTested - how many samples (wells) with that cpd have been tested
   wActive - how many samples (wells) with that cpd have been tested active
 
 Output (2): activity file, each line:
   CID SID AID Outcome
     where:
   1 = inactive
   2 = active
   3 = inconclusive
   4 = unspecified
   5 = probe
   multiple, differing 1, 2 or 3 = discrepant
   not 4 = tested
"""
  aids_tested=set(); aids_active=set();
  sids_tested=set(); sids_active=set();
  n_sam=0; n_sam_active=0; mol_active=False; mol_found=False;

  try:
    fcsv = requests.get(base_url+f"/compound/cid/{cid}/assaysummary/CSV").text
  except Exception as e:
    logging.error(f"[{cid}] REST request failed; {e}")
    fout_mol.write(f"{smiles}\t{cid}\n")
    return mol_found,mol_active,n_sam,n_sam_active
  mol_found=True

  try:
    csvReader=csv.DictReader(fcsv.splitlines(),fieldnames=None,restkey=None,restval=None,dialect='excel',delimiter=',',quotechar='"')
    csvrow = csvReader.next()    ## must do this to get fieldnames
  except Exception as e:
    logging.error(f"[CID={cid}] CSV problem:{e}")
    fout_mol.write(f"{smiles}\t{cid}\n")
    return mol_found,mol_active,n_sam,n_sam_active

  for field in ('AID','CID','SID','Bioactivity Outcome','Bioassay Type'):
    if field not in csvReader.fieldnames:
      logging.error(f"[CID={cid}] bad CSV header, no '{field}' field.")
      logging.debug("fieldnames: "+(','.join(map((lambda x:'"%s"'%x),csvReader.fieldnames))))
      fout_mol.write(f"{smiles}\t{cid}\n")
      return mol_found,mol_active,n_sam,n_sam_active

  n_in=0
  while True:
    try:
      csvrow=csvReader.next()
    except:
      break  ## EOF
    n_in+=1

    try:
      aid=int(csvrow['AID'])
      sid=int(csvrow['SID'])
      csvrow_cid=int(csvrow['CID'])
    except:
      logging.error(f"[CID={cid}] bad CSV line; problem parsing: '{str(csvrow)}'")
      continue
    if cid!=csvrow_cid:
      logging.error(f"Aaack! [CID={cid}] CID mismatch: != {csvrow_cid}")
      return mol_found,mol_active,n_sam,n_sam_active

    if aidset:
      if aid not in aidset:
        continue
      #logging.debug(f"AID [{aid}] ok (pre-selected).")

    ## Assay filtering done; now update statistics (sTested, sActive).
    n_sam+=1
    aidhash[aid]=True
    act_outcome=csvrow['Bioactivity Outcome'].lower()
    if act_outcome in OUTCOME_CODES:
      fout_act.write(f"{cid}\t{sid}\t{aid}\t{OUTCOME_CODES[act_outcome]}\n")
    else:
      logging.debug(f"[{n_in}] unrecognized outcome (CID={cid},AID={aid}): '{act_outcome}'")
    aids_tested.add(aid)
    sids_tested.add(sid)
    if act_outcome=='active':
      n_sam_active+=1
      mol_active=True
      aids_active.add(aid)
      sids_active.add(sid)

  logging.debug(f"[CID={cid}] CSV data lines: {n_in}")
  fout_mol.write(f"{smiles}\t{cid}\t{len(sids_tested)}\t{len(sids_active)}\t{len(aids_tested)}\t{len(aids_active)}\t{n_sam}\t{n_sam_active}\n")
  logging.debug("cid={cid},sTested={len(sids_tested)},sActive={len(sids_active)},aTested={len(aids_tested)},aActive={len(aids_active)},wTested={n_sam},wActive={n_sam_active}")
  return mol_found,mol_active,n_sam,n_sam_active

#############################################################################
def GetCpdAssayData(cid_query, aidset, base_url=BASE_URL, fout=None):
  """This function needs fixing."""
  try:
    fcsv = requests.get(base_url+f"/compound/cid/{cid_query}/assaysummary/CSV").text
  except Exception as e:
    logging.error(f"[{cid_query}] REST request failed; {e}")
    return False

  if not fcsv:
    return False

  try:
    csvReader=csv.DictReader(fcsv.splitlines(),fieldnames=None,restkey=None,restval=None,dialect='excel',delimiter=',',quotechar='"')
    csvrow=csvReader.next()    ## must do this to get fieldnames
  except Exception as e:
    logging.error(f"[CID={cid_query}] CSV problem: {e}")
    return True

  for field in ('CID','SID','AID','Bioactivity Outcome','Activity Value [uM]'):
    if field not in csvReader.fieldnames:
      logging.error(f"[CID={cid_query}] bad CSV header, no '{field}' field.")
      logging.debug("fieldnames: "+(','.join(map((lambda x:'"%s"'%x),csvReader.fieldnames))))
      return True

  n_in=0; n_act=0;
  while True:
    try:
      csvrow=csvReader.next()
    except:
      break  ## EOF
    n_in+=1
    try:
      aid=int(csvrow['AID'])
      sid=int(csvrow['SID'])
      cid=int(csvrow['CID'])
      outcome=csvrow['Bioactivity Outcome']
    except:
      logging.error(f"[CID={cid}] bad CSV line; problem parsing.")
      logging.debug(f"csvrow = {str(csvrow)}")
      continue
    if cid_query!=cid:
      logging.error(f"Aaack! [CID={cid}] CID mismatch: != {int(csvrow['CID'])}")
      return True
    if aid not in aidset:
      continue
      logging.debug(f"AID [{aid}] ok (pre-selected).")

    if not csvrow['Activity Value [uM]']:
      continue

    try:
      actval=float(csvrow['Activity Value [uM]'])
    except:
      logging.error(f"[CID={cid}] bad CSV line; problem parsing activity: '{csvrow['Activity Value [uM]']}'")
      continue

    n_act+=1
    outcome_code = OUTCOME_CODES[outcome.lower()] if outcome.lower() in OUTCOME_CODES else 0

    fout.write(f"{cid}\t{sid}\t{aid}\t{outcome_code}\t{actval:.3f}\n")

  logging.info(f"[CID={cid_query}] records: {n_in}; activities: {n_act}")
  return True

#############################################################################



#!/usr/bin/env python3
##############################################################################
### uniprot_utils.py - utility functions for access to Uniprot REST API.
### UniprotKB = Uniprot Knowledge Base
##############################################################################
import sys,os,re,logging
#
from ..util import rest
#
#############################################################################
def GetData(base_uri, uids, ofmt, fout):
  """Need to handle xml, rdf better (merge)."""
  n_prot=0; n_err=0;
  for uid in uids:
    rval=rest.GetURL(base_uri+'/%s.%s'%(uid, ofmt))
    if not rval:
      n_err+=1
      continue
    if ofmt=='tab':
      lines=[];
      for line in rval.splitlines():
        vals=re.split(r'\t', line)
        lines.append('\t'.join(vals))
      for i,line in enumerate(lines):
        if n_prot>0 and i==0: continue #Skip duplicate headers.
        fout.write(line+'\n')
    else:
      fout.write(rval+'\n')
    n_prot+=1
  logging.info('n_in: %d; n_prot: %d; n_err: %d'%(len(uids), n_prot, n_err))

#############################################################################
def UIDs2JSON(base_uri, uids, fout):
  """ Uses uniprot library from Bosco Ho (https://github.com/boscoh/uniprot)."""
  import uniprot ## Bosco Ho (https://github.com/boscoh/uniprot)
  uniprot_data=uniprot.batch_uniprot_metadata(uids, None)
  for uid in uniprot_data.keys():
    for key in uniprot_data[uid].keys():
      if key in ('accs', 'sequence', 'go', 'description'):  #keep simple
        del uniprot_data[uid][key]
  json_txt=json.dumps(uniprot_data, sort_keys=True, indent=2)
  fout.write(json_txt+'\n')
  return



#!/usr/bin/env python3
"""
"""
import sys,os,re,time,json,logging,requests,tqdm
import urllib,urllib.parse
import pandas as pd
#
RESOLVER_API_HOST='resolver.api.identifiers.org'
RESOLVER_API_BASE_PATH=''
RESOLVER_API_BASE_URL='https://'+RESOLVER_API_HOST+RESOLVER_API_BASE_PATH
#
REGISTRY_API_HOST='registry.api.identifiers.org'
REGISTRY_API_BASE_PATH='/restApi'
REGISTRY_API_BASE_URL='https://'+REGISTRY_API_HOST+REGISTRY_API_BASE_PATH
#
NCHUNK=100
#
##############################################################################
def Resolve(ids, base_url=RESOLVER_API_BASE_URL, fout=None):
  n_out=0; tags=None; df=None;
  for id_this in ids:
    response = requests.get(f"{base_url}/{id_this}")
    logging.debug(response.text)
    results = response.json()
    resources = results['payload']['resolvedResources']
    for resource in resources:
      if not tags:
        tags = list(resource.keys())
        for tag in tags[:]:
          if type(resource[tag]) in (list, dict):
            logging.info(f"Ignoring field: {tag}")
            tags.remove(tag)
      df_this = pd.DataFrame({tag:[resource[tag]] for tag in tags})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  return df

##############################################################################
def ListEntities(entityname, base_url=REGISTRY_API_BASE_URL, fout=None):
  n_out=0; tags=None; df=None;
  page=0; size=NCHUNK;
  while True:
    url_this = f"{base_url}/{entityname}s?page={page}&size={size}"
    response = requests.get(url_this)
    logging.debug(response.text)
    if response.status_code!=200: break
    results = response.json()
    things = results['_embedded'][f"{entityname}s"]
    if not things: break
    for thing in things:
      if not tags:
        tags = list(thing.keys())
        for tag in tags[:]:
          if type(thing[tag]) in (list, dict):
            logging.info(f"Ignoring field: {tag}")
            tags.remove(tag)
      df_this = pd.DataFrame({tag:[thing[tag]] for tag in tags})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
    page+=1
  logging.info(f"n_out: {n_out}")
  return df

#############################################################################
def ListNamespaces(base_url=REGISTRY_API_BASE_URL, fout=None):
  return ListEntities("namespace", base_url, fout)

def ListResources(base_url=REGISTRY_API_BASE_URL, fout=None):
  return ListEntities("resource", base_url, fout)

def ListInstitutions(base_url=REGISTRY_API_BASE_URL, fout=None):
  return ListEntities("institution", base_url, fout)

##############################################################################
def SearchEntities(query, search_logic, entityname, base_url=REGISTRY_API_BASE_URL, fout=None):
  n_out=0; tags=None; df=None;
  page=0; size=NCHUNK;
  url = f"{base_url}/{entityname}s/search/findByName?name={urllib.parse.quote(query)}" if search_logic=="exact" else f"{base_url}/{entityname}s/search/findByNameContaining?nameContent={urllib.parse.quote(query)}"
  while True:
    url_this = f"{url}&page={page}&size={size}"
    response = requests.get(url_this)
    logging.debug(response.text)
    if response.status_code!=200: break
    results = response.json()
    things = results['_embedded'][f"{entityname}s"]
    if not things: break
    for thing in things:
      if not tags:
        tags = list(thing.keys())
        for tag in tags[:]:
          if type(thing[tag]) in (list, dict):
            logging.info(f"Ignoring field: {tag}")
            tags.remove(tag)
      df_this = pd.DataFrame({tag:[thing[tag]] for tag in tags})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, "\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
    page+=1
  logging.info(f"n_out: {n_out}")
  return df

##############################################################################
def SearchNamespaces(query, search_logic, base_url=REGISTRY_API_BASE_URL, fout=None):
  return SearchEntities(query, search_logic, "namespace", base_url, fout)

def SearchInstitutions(query, search_logic, base_url=REGISTRY_API_BASE_URL, fout=None):
  return SearchEntities(query, search_logic, "institution", base_url, fout)

##############################################################################



#!/usr/bin/env python3
##############################################################################
### wikipathways_utils.py - utility for Wikipathways REST API.
### See: http://www.wikipathways.org/index.php/Help:WikiPathways_Webservice/API
### Formats: xml, json
##############################################################################
import sys,os,re,argparse,time,logging
#
from .. import wikipathways
#
##############################################################################
if __name__=='__main__':
  OFMTS = ["gpml", "png", "svg", "pdf", "txt", "pwf", "owl"]
  API_HOST='webservice.wikipathways.org'
  API_BASE_PATH=''
  search_params = {'human':False}
  ops = ['list_organisms', 'list_pathways', 'get_pathway']
  parser = argparse.ArgumentParser(description='WikiPathways REST API client')
  parser.add_argument("op", choices=ops, help='OPERATION (select one)')
  parser.add_argument("--ids", help="input IDs")
  parser.add_argument("--i", dest="ifile", help="input file, IDs")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--skip", type=int, default=0)
  parser.add_argument("--nmax", type=int, default=None)
  parser.add_argument("--nchunk", type=int, default=None)
  parser.add_argument("--regex", help="regular expression query")
  parser.add_argument("--str_query", help="string query")
  parser.add_argument("--ofmt", choices=OFMTS, default="gpml", help="output pathway image format")
  parser.add_argument("--human", action="store_true", help="human only")
  parser.add_argument("--api_host", default=API_HOST)
  parser.add_argument("--api_base_path", default=API_BASE_PATH)
  parser.add_argument("-v","--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  fout = open(args.ofile, 'w') if args.ofile else sys.stdout

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
  elif args.ids:
    ids = re.split('[, ]+', args.ids.strip())
  if len(ids)>0: logging.info('Input IDs: %d'%(len(ids)))

  api_base_url='https://'+args.api_host+args.api_base_path

  t0=time.time()

  if args.op=="list_organisms":
    wikipathways.ListOrganisms(api_base_url, fout)

  elif args.op=="list_pathways":
    wikipathways.ListPathways(api_base_url, search_params, fout)

  elif args.op=="get_pathway":
    wikipathways.GetPathway(api_base_url, ids, ofmt, fout)

  else:
    parser.error('Invalid operation: {0}'.format(args.op))

  logging.info('Elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0))))




#!/usr/bin/env python
#############################################################################
### pug_substance_search.py - PubChem/PUG substance search
### Jeremy Yang
###   4 Feb 2009
#############################################################################
import sys,os,re,urllib,time,getopt,tempfile
import pug_utils

PROG=os.path.basename(sys.argv[0])

PUGURL='http://pubchem.ncbi.nlm.nih.gov/pug/pug.cgi'
POLL_WAIT=10
MAX_WAIT=300
SIM_CUTOFF=0.80
NMAX=0

#############################################################################
def QueryPug_qkey(qkey,webenv,fmt='smiles',gzip=False,verbose=0):
  if gzip: compres='gzip'
  else: compres='none'
  qxml='''\
<PCT-Data><PCT-Data_input><PCT-InputData>
  <PCT-InputData_download><PCT-Download> 
    <PCT-Download_uids><PCT-QueryUids>
      <PCT-QueryUids_entrez><PCT-Entrez>
        <PCT-Entrez_db>pcsubstance</PCT-Entrez_db>
        <PCT-Entrez_query-key>%(QKEY)s</PCT-Entrez_query-key>
        <PCT-Entrez_webenv>%(WEBENV)s</PCT-Entrez_webenv>
      </PCT-Entrez></PCT-QueryUids_entrez>
    </PCT-QueryUids></PCT-Download_uids>
    <PCT-Download_format value="%(FMT)s"/>
    <PCT-Download_compression value="%(COMPRES)s"/>
  </PCT-Download></PCT-InputData_download>
</PCT-InputData></PCT-Data_input></PCT-Data>
'''%{'QKEY':qkey,'WEBENV':webenv,'FMT':fmt,'COMPRES':compres}
  if verbose:
    print >>sys.stderr,'connecting %s...'%PUGURL
    print >>sys.stderr,'submitting query for qkey: %s ...'%qkey
  f=pug_utils.UrlOpenTry(PUGURL,qxml)
  pugxml=f.read()
  f.close()
  status,reqid,url,qkey,wenv,error=pug_utils.ParsePugXml(pugxml)
  return status,reqid,url,qkey,wenv,error

#############################################################################
def QueryPug_struct(smiles,sim=False,sim_cutoff=0.8,active=False,mlsmr=False,nmax=10000,verbose=0):
  qxml='''\
<PCT-Data><PCT-Data_input><PCT-InputData><PCT-InputData_query><PCT-Query>
  <PCT-Query_type><PCT-QueryType><PCT-QueryType_css>
    <PCT-QueryCompoundCS>
    <PCT-QueryCompoundCS_query>
      <PCT-QueryCompoundCS_query_data>%(SMILES)s</PCT-QueryCompoundCS_query_data>
    </PCT-QueryCompoundCS_query>
    <PCT-QueryCompoundCS_type>
'''%{'SMILES':smiles}
  if sim:
    qxml+='''\
      <PCT-QueryCompoundCS_type_similar><PCT-CSSimilarity>
        <PCT-CSSimilarity_threshold>%(SIM_CUTOFF)d</PCT-CSSimilarity_threshold>
      </PCT-CSSimilarity></PCT-QueryCompoundCS_type_similar>
'''%{'SIM_CUTOFF':sim_cutoff*100}
  else:
    qxml+='''\
      <PCT-QueryCompoundCS_type_subss><PCT-CSStructure>
        <PCT-CSStructure_isotopes value="false"/>
        <PCT-CSStructure_charges value="false"/>
        <PCT-CSStructure_tautomers value="false"/>
        <PCT-CSStructure_rings value="false"/>
        <PCT-CSStructure_bonds value="true"/>
        <PCT-CSStructure_chains value="true"/>
        <PCT-CSStructure_hydrogen value="false"/>
      </PCT-CSStructure></PCT-QueryCompoundCS_type_subss>
'''
  qxml+='''\
    </PCT-QueryCompoundCS_type>
    <PCT-QueryCompoundCS_results>%(NMAX)d</PCT-QueryCompoundCS_results>
    </PCT-QueryCompoundCS>
  </PCT-QueryType_css></PCT-QueryType>
'''%{'NMAX':nmax}
  if active:
    qxml+='''\
  <PCT-QueryType><PCT-QueryType_cel><PCT-QueryCompoundEL>
    <PCT-QueryCompoundEL_activity>
      <PCT-CLByBioActivity value="active">2</PCT-CLByBioActivity>
    </PCT-QueryCompoundEL_activity>
  </PCT-QueryCompoundEL></PCT-QueryType_cel></PCT-QueryType>
'''
  if mlsmr:
    qxml+='''\
  <PCT-QueryType><PCT-QueryType_cel><PCT-QueryCompoundEL>
    <PCT-QueryCompoundEL_source><PCT-CLByString>
      <PCT-CLByString_qualifier value="must">1</PCT-CLByString_qualifier>
      <PCT-CLByString_category>MLSMR</PCT-CLByString_category>
    </PCT-CLByString></PCT-QueryCompoundEL_source>
  </PCT-QueryCompoundEL></PCT-QueryType_cel></PCT-QueryType>
'''
  qxml+='''\
  </PCT-Query_type></PCT-Query>
</PCT-InputData_query></PCT-InputData></PCT-Data_input></PCT-Data>
'''
  if verbose:
    print >>sys.stderr,'connecting %s...'%PUGURL
    if sim:
      print >>sys.stderr,'query (%f): %s ...'%(sim_cutoff,smiles)
    else:
      print >>sys.stderr,'substructure query: %s ...'%smiles
  f=pug_utils.UrlOpenTry(PUGURL,qxml)
  pugxml=f.read()
  f.close()

  status,reqid,url,qkey,wenv,error=pug_utils.ParsePugXml(pugxml)
  return status,reqid,qkey,wenv,error

#############################################################################
def ElapsedTime(t_start):
  t_elapsed=time.time()-t_start
  tsec=t_elapsed%60
  tmin=t_elapsed/60
  return ("%d:%02d"%(tmin,tsec))

#############################################################################
if __name__=='__main__':
  usage='''
  %(PROG)s - PubChem substructure or similarity search, fetch smiles

  required:
  --name_substr=<str>   ... name substring query
  --o=<OUTSMIS> ... output smiles file (w/ CIDs)
  --fmt=<FMT> ... smiles|sdf [default via filename]
  --gz           ... output gzipped [default via filename]
  --out_sids=<OUTSIDS> ... output SIDs (only w/ smiles output)
  --active            ... active mols only (in any assay)
  --mlsmr             ... MLSMR mols only
  --nmax            ... max hits returned
  --max_wait     ... max wait for query
  --v            ... verbose
  --vv           ... very verbose
  --h            ... this help
'''%{'PROG':PROG,'SIM_CUTOFF':SIM_CUTOFF}

  def ErrorExit(msg):
    print >>sys.stderr,msg
    sys.exit(1)

  smiles=None; ofile=None; verbose=0;
  out_cids_file=None; similarity=False; active=False; mlsmr=False;
  fmt=None; gzip=False;
  opts,pargs=getopt.getopt(sys.argv[1:],'',['h','v','vv','smi=','smiles=',
  'o=','out=','nmax=','n=','sim_cutoff=','similarity','sim','out_cids=',
  'fmt=','gz','max_wait=','active','mlsmr'])
  if not opts: ErrorExit(usage)
  for (opt,val) in opts:
    if opt=='--h': ErrorExit(usage)
    elif opt in ('--smiles','--smi'): smiles=val
    elif opt in ('--o','--out'): ofile=val
    elif opt in ('--fmt','--format'): fmt=val
    elif opt=='--out_cids': out_cids_file=val
    elif opt in ('--similarity','--sim'): similarity=True
    elif opt=='--active': active=True
    elif opt=='--mlsmr': mlsmr=True
    elif opt=='--sim_cutoff': SIM_CUTOFF=float(val)
    elif opt=='--max_wait': MAX_WAIT=int(val)
    elif opt in ('--n','--nmax'): NMAX=int(val)
    elif opt=='--vv': verbose=2
    elif opt=='--v': verbose=1
    else: ErrorExit('Illegal option: %s'%val)

  if not smiles:
    ErrorExit('smiles required\n'+usage)
  if verbose:
    print >>sys.stderr,'query: "%s"'%smiles

  if ofile:
    fout=file(ofile,'wb')
    if not fout:
      ErrorExit('cannot open: %s\n%s'%(ofile,usage))
    if (ofile[-3:]=='.gz'):
      gzip=True
      ext=re.sub('^.*\.','',ofile[:-3])
    else:
      ext=re.sub('^.*\.','',ofile)
    if not fmt: fmt=ext
    if fmt=='smi': fmt='smiles'
    if fmt not in ('smiles','sdf'):
      ErrorExit('format "smiles" or "sdf" required\n'+usage)

  if out_cids_file:
    if fmt=='sdf':
      ErrorExit('--out_cids not allowed with SDF output.\n'+usage)

    fout_cids=file(out_cids_file,'wb')
    if not fout_cids:
      ErrorExit('cannot open: %s\n%s'%(out_cids_file,usage))

  t0=time.time()	## start timer

  status,reqid,qkey,wenv,error=QueryPug_struct(smiles,similarity,SIM_CUTOFF,active,mlsmr,NMAX,verbose)

  if status not in ('success','queued','running'):
    ErrorExit('query failed; quitting (status=%s,error="%s").'%(status,error))
  if verbose:
    print >>sys.stderr,'query status: %s'%status

  if not reqid and not qkey:
    ErrorExit('query ok but no ID; quitting.')

  i_poll=0
  while not qkey:
    if time.time()-t0>MAX_WAIT:
      pug_utils.PollPug(reqid,'cancel',verbose)
      ErrorExit('max wait exceeded (%d sec); quitting.'%MAX_WAIT)
    print >>sys.stderr,'polling PUG [ID=%d]...'%reqid
    status,reqid_new,url,qkey,wenv,error=pug_utils.PollPug(reqid,'status',verbose)
    if reqid_new: reqid=reqid_new
    if qkey: break
    if verbose:
      print >>sys.stderr,'%s elapsed; %d sec wait; status=%s'%(ElapsedTime(t0),POLL_WAIT,status)
    print >>sys.stderr,'%s'%(error)
    if re.search('No result found',error,re.I):
      ErrorExit(error)
    time.sleep(POLL_WAIT)
    i_poll+=1


  status,reqid,url,qkey,wenv,error=QueryPug_qkey(qkey,wenv,fmt,gzip,verbose)
  i_poll=0
  while not url:
    if time.time()-t0>MAX_WAIT:
      pug_utils.PollPug(reqid,'cancel',verbose)
      ErrorExit('max wait exceeded (%d sec); quitting.'%MAX_WAIT)
    print >>sys.stderr,'polling PUG [ID=%d]...'%reqid
    status,reqid,url,qkey,wenv,error=pug_utils.PollPug(reqid,'status',verbose)
    if url: break
 
    if verbose:
      print >>sys.stderr,'%s elapsed; %d sec wait; status=%s'%(ElapsedTime(t0),POLL_WAIT,status)
    time.sleep(POLL_WAIT)
    i_poll+=1

  if verbose:
    print >>sys.stderr,'query elapsed time: %s'%(ElapsedTime(t0))
  if ofile:
    if verbose:
      print >>sys.stderr,'URL: %s'%url
      print >>sys.stderr,'downloading to %s...'%ofile

  fd,tmpfile=tempfile.mkstemp('.txt',PROG)
  f=os.fdopen(fd,"w")
  nbytes=pug_utils.DownloadUrl(url,f)
  f.close()

  if fmt=='smiles':
    ftmp=open(tmpfile)
    n=0
    while True:
      line=ftmp.readline()
      if not line: break
      fields=line.split()
      if ofile: fout.write("%s %s\n"%(fields[1],fields[0]))
      if out_cids_file: fout_cids.write("%s\n"%(fields[0]))
      n+=1
    print >>sys.stderr,'compounds: %d'%(n)
    ftmp.close()
  else:
    os.system("cp %s %s"%(tmpfile,ofile))
  os.unlink(tmpfile)

  sys.stderr.write('%s: format: %s'%(PROG,fmt))
  if gzip: sys.stderr.write('(gzipped)')
  sys.stderr.write('\n')

  if ofile:
    fout.close()
    nbytes=os.stat(ofile).st_size
    print >>sys.stderr,'%s (%.2fMB)'%(ofile,nbytes/1e6)

  if out_cids_file:
    fout_cids.close()
    nbytes=os.stat(out_cids_file).st_size
    print >>sys.stderr,'%s (%.2fMB)'%(out_cids_file,nbytes/1e6)




#!/usr/bin/env python3
"""
Access to Ensembl BIOMART REST API.
https://m.ensembl.org/info/data/biomart/biomart_restful.html
"""
import sys,os,re,argparse,time,logging

from ... import ensembl
#
##############################################################################
def DemoXMLQuery(base_url, fout):
  xmltext = """\
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE Query>
<Query  virtualSchemaName = "default" formatter = "TSV" header = "1" uniqueRows = "0" count = "" datasetConfigVersion = "0.6" >

	<Dataset name = "hsapiens_gene_ensembl" interface = "default" >
		<Attribute name = "ensembl_gene_id" />
		<Attribute name = "ensembl_gene_id_version" />
		<Attribute name = "entrezgene_id" />
		<Attribute name = "hgnc_id" />
		<Attribute name = "hgnc_symbol" />
	</Dataset>
</Query>
"""
  ensembl.biomart.Utils.XMLQuery(xmltext, base_url, fout)

##############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(prog=sys.argv[0], description="Ensembl BIOMART REST API client", epilog="For XML query file format, see https://m.ensembl.org/info/data/biomart/biomart_restful.html#biomartxml")
  ops = ["xmlQuery", "ensg2ncbi", "ensg2hgnc", "ensg2ncbihgnc", "demo", "show_version"]
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--ixml", dest="ixmlfile", help="input file, XML query")
  parser.add_argument("--api_host", default=ensembl.biomart.API_HOST)
  parser.add_argument("--api_base_path", default=ensembl.biomart.API_BASE_PATH)
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("-v", "--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url='http://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w") if args.ofile else sys.stdout

  t0=time.time()

  if args.op=='xmlQuery':
    if not args.ixmlfile:
      parser.error(f"Input XML file required for {args.op}")
      sys.exit(1)
    xmltext = open(args.ixmlfile).read()
    ensembl.biomart.Utils.XMLQuery(xmltext, base_url, fout)

  elif args.op=='ensg2ncbi':
    ensembl.biomart.Utils.ENSG2NCBI(base_url, fout)

  elif args.op=='ensg2hgnc':
    ensembl.biomart.Utils.ENSG2HGNC(base_url, fout)

  elif args.op=='ensg2ncbihgnc':
    ensembl.biomart.Utils.ENSG2NCBIHGNC(base_url, fout)

  elif args.op=='demo':
    DemoXMLQuery(base_url, fout)

  else:
    parser.error(f'Invalid operation: {args.op}')

  logging.info(('%s: elapsed time: %s'%(os.path.basename(sys.argv[0]), time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0)))))



#!/usr/bin/env python3
###
# https://mygene.info/
# https://pypi.org/project/mygene/
###
import sys,os,re,argparse,time,logging
import pandas as pd
import mygene as mg
#
from .. import mygene as bc_mygene
#
#############################################################################
if __name__=='__main__':
  epilog = "See https://mygene.info/, https://pypi.org/project/mygene/.  Example queries: 'cdk2', 'symbol:cdk2', 'symbol:cdk*'"
  ops = ["get", "search"]
  parser = argparse.ArgumentParser(description='MyGene API client', epilog=epilog)
  parser.add_argument("op", choices=ops, help="OPERATION")
  parser.add_argument("--i", dest="ifile", help="input gene IDs or queries")
  parser.add_argument("--ids", help="input gene IDs or queries, comma-separated")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--species", default="human", help="species name or taxonomy ID")
  parser.add_argument("--fields", default=bc_mygene.FIELDS, help="requested fields")
  parser.add_argument("-v", "--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  fout = open(args.ofile, 'w') if args.ofile else sys.stdout

  t0 = time.time()
  logging.info('Python: {}; mygene: {}'.format(sys.version.split()[0], mg.__version__))

  if args.ifile:
    genes = pd.read_table(args.ifile, header=None, names=["ID"])
    ids = list(genes.ID)
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids)
  else:
    parser.error("Input IDs required via --i or --ids.")

  fields = re.split(r'[,\s]+', args.fields)

  if args.op=="get":
    bc_mygene.GetGenes(ids, fields, fout)

  elif args.op=="search":
    bc_mygene.SearchGenes(ids, args.species, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")

  logging.info(('elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0)))))




#!/usr/bin/env python3
#############################################################################
### For processing PubChem FTP files.
### 
### PUBCHEM_ACTTIVITY_OUTCOME values:
###  Inactive:1, Active:2, Inconclusive:3, Unspecified:4, Probe:5
#############################################################################
import sys,os,re,time,csv,logging
import urllib,urllib.request,tempfile

OUTCOME_CODES = {'inactive':1,'active':2,'inconclusive':3,'unspecified':4,'probe':5}

#############################################################################
def UrlOpen(url, ntries=20, poll_wait=10):
  ### To recover from FTP timeouts:
  fin=None
  for i in range(ntries):
    try:
      req = urllib.request.Request(url=url)
      fin = urllib.request.urlopen(req)
      break
    except IOError as e:
      logging.error('IOError: %s'%e)
      time.sleep(poll_wait)
  if not fin:
    logging.error('failed download, %d tries, quitting...'%ntries)
    sys.exit(1)
  return fin

#############################################################################
def GetUrl(url,fout,ntries=20,poll_wait=10):
  logging.info('%s'%url)
  fin=UrlOpen(url,ntries,poll_wait)
  nbytes=0
  while True:
    buff=fin.read(1024)
    if not buff: break
    fout.write(buff)
    nbytes+=len(buff)
  fin.close()
  return nbytes

#############################################################################
def GetUrlSDF2SMI(url,fout,ntries=20,poll_wait=10):
  '''Get PubChem SDF.GZ, convert to SMILES using the SD tag PUBCHEM_OPENEYE_CAN_SMILES
or PUBCHEM_OPENEYE_ISO_SMILES, and PUBCHEM_COMPOUND_CID or PUBCHEM_COMPOUND_SID for name.'''
  import openeye.oechem as oechem

  def HandleOEErrors(oeerrs,nowarn):
    errstr=oeerrs.str()
    for line in errstr.split('\n'):
      if not line.rstrip(): continue
      if re.search('Warning',line,re.I) and nowarn: continue
      sys.stderr.write("%s\n"%line)
    oeerrs.clear()

  fout_tmp = tempfile.NamedTemporaryFile(prefix='pubchem_ftp_',suffix='.sdf.gz',delete=False)
  GetUrl(url,fout_tmp,ntries,poll_wait)
  fpath_tmp=fout_tmp.name
  logging.debug('fpath_tmp = %s'%fpath_tmp)
  fout_tmp.close()

  ims=oechem.oemolistream(fpath_tmp)
  ims.SetFormat(oechem.OEFormat_SDF)
  ims.Setgz(True)

  mol=oechem.OEGraphMol()
  nbytes=0
  oeerrs=oechem.oeosstream()
  oechem.OEThrow.SetOutputStream(oeerrs)
  while oechem.OEReadMolecule(ims,mol):
    cid = oechem.OEGetSDData(mol,'PUBCHEM_COMPOUND_CID')
    cansmi = oechem.OEGetSDData(mol,'PUBCHEM_OPENEYE_CAN_SMILES')
    isosmi = oechem.OEGetSDData(mol,'PUBCHEM_OPENEYE_ISO_SMILES')
    buff=("%s %s\n"%(isosmi,cid))
    fout.write(buff)
    nbytes+=len(buff)
    HandleOEErrors(oeerrs,True)
  os.remove(fpath_tmp)
  return nbytes


#############################################################################
### ExtractOutcomes() - return hash of all SIDs/CIDs to list of
### activity values [1-4] (may be multiple and discrepant).
### Also return hash of outcomes from this assay.
### If sidset exists this is a subset of interest.
### If use_cids is true, all ids are CIDs.
#############################################################################
def ExtractOutcomes(ftxt,sidset,use_cids=False):
  sids={} 
  sids_inactive=[]; sids_active=[]; sids_inconclusive=[];
  sids_unspecified=[]; sids_discrepant=[];
  lines=ftxt.splitlines()
  if len(lines)<2:
    return sids
  logging.debug('tag line: %s'%(lines[0]))
  if use_cids:
    id_tag='PUBCHEM_CID'
  else:
    id_tag='PUBCHEM_SID'
  tags=lines[0].split(',')
  j_sid,j_cid,j_act=None,None,None
  for j,tag in enumerate(tags):
    if re.match('"?%s"?$'%id_tag,tag): j_sid=j
    if re.match('"?PUBCHEM_ACTIVITY_OUTCOME"?$',tag): j_act=j
  if j_sid==None or j_act==None:
    logging.info('cannot find sid and activity tags: %s'%(lines[0]))
    return sids
  for line in lines[1:]:
    try:
      vals=line.split(',')
      ## sid may be absent; report?
      sid=re.sub('"','',vals[j_sid])
      if not sid: continue
      sid=int(sid)
      act=int(vals[j_act])	## should be [1-4]
    except:
      logging.warning('Problem with line: %s'%(line))
      #break	#DEBUG
      continue
    if sidset and sid not in sidset : continue

    if sid in sids:
      sids[sid]['acts'].append(act)
    else:
      sids[sid]={'acts':[act],'outcome':0}

  for sid in sids.keys():
    acts=sids[sid]['acts']
    acts.sort()
    while 4 in acts and len(acts)>1: acts.remove(4)
    if len(acts)==1 or acts[0]==acts[-1]:
      sids[sid]['outcome']=acts[0]
    else:
      sids[sid]['outcome']=5
      if not ( (1 in acts and 2 in acts)
        or (1 in acts and 3 in acts)
        or (2 in acts and 3 in acts)): logging.error('Doh!  acts=%s'%(str(acts)))

  return sids
    
#############################################################################
### Input file is assay CSV.  For specified SIDs, output CSV with outcomes.
### 4 header lines expected.  Maybe 10 fields.
#############################################################################
def ExtractResults(fin,aid,sids,fout):
  n_in=0; n_out=0; n_active=0;
  tags = None;
  j_sid = None; j_actout = None; j_actsco = None;
  fout.write('AID,SID,ACTIVITY_OUTCOME\n')
  while True:
    line = fin.readline()
    if not line: break
    n_in+=1
    if n_in==1 or not tags:
      tags = re.split(',',line)
      logging.debug('DEBUG: %s'%line)
      for j,tag in enumerate(tags):
        logging.debug('\t"%s"'%tag)
        if tag == 'PUBCHEM_SID': j_sid = j
        elif tag == 'PUBCHEM_ACTIVITY_OUTCOME': j_actout = j
        elif tag == 'PUBCHEM_ACTIVITY_SCORE': j_actsco = j
      continue
    elif n_in<5: continue
    if not j_sid:
      logging.error('PUBCHEM_SID column not found.')
      break
    if not j_actout:
      logging.error('PUBCHEM_ACTIVITY_OUTCOME column not found.')
      break
    vals = re.split(',',line.rstrip())
    sid = vals[j_sid]
    if not sid: continue
    sid = int(sid)
    if sid not in sids: continue
    actout = vals[j_actout]
    if actout=='Active': n_active+=1
    if actout.lower() in OUTCOME_CODES:
      actout = OUTCOME_CODES[actout.lower()]
    fout.write('%d,%d,%s\n'%(aid,sid,actout))
    n_out+=1

  logging.info('%s: n_in = %d ; n_out = %d ; n_active = %d'%(aid,n_in,n_out,n_active))

#############################################################################
def Str2Ints(str):
  if not re.search('\S',str): return []
  return map(lambda x:int(x),str.split(','))

#############################################################################
def Ints2Str(intlist):
  if not intlist: return ''
  return (','.join(map(lambda x:('%d'%x),intlist)))

#############################################################################
